# Parameters whose values may vary.

VARIABLE_PARAMS = set(
    [
        "effective_search_space",
        "effective_hsp_length",
        "query",
        "query_length",
        "query_letters",
    ]
)


def checkCompatibleParams(initialParams, laterParams):
    """
    Check a later set of BLAST parameters against those originally found.

    @param initialParams: A C{dict} with the originally encountered BLAST
        parameter settings.
    @param laterParams: A C{dict} with BLAST parameter settings encountered
        later.
    @return: A C{str} summary of the parameter differences if the parameter
        sets differ, else C{None}.
    """

    # Note that although the params contains a 'date', its value is empty
    # (as far as I've seen). This could become an issue one day if it
    # becomes non-empty and differs between JSON files that we cat
    # together. In that case we may need to be more specific in our params
    # compatible checking.
    err = []
    for param in initialParams:
        if param in laterParams:
            if (
                param not in VARIABLE_PARAMS
                and initialParams[param] != laterParams[param]
            ):
                err.append(
                    "\tParam %r initial value %r differs from "
                    "later value %r" % (param, initialParams[param], laterParams[param])
                )
        else:
            err.append(
                "\t%r found in initial parameters, not found "
                "in later parameters" % param
            )
    for param in laterParams:
        if param not in initialParams:
            err.append(
                "\t%r found in later parameters, not seen in "
                "initial parameters" % param
            )

    return "Summary of differences:\n%s" % "\n".join(err) if err else None



import os
from copy import deepcopy
from stat import S_ISDIR
from math import ceil
from collections import defaultdict
from time import ctime, time
from textwrap import fill

import matplotlib

if not os.environ.get("DISPLAY"):
    # Use non-interactive Agg backend
    matplotlib.use("Agg")

import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib import gridspec, patches

import numpy as np

from dark.aa import propertiesForSequence, clustersForSequence
from dark.baseimage import BaseImage
from dark.dimension import dimensionalIterator
from dark.html import AlignmentPanelHTMLWriter, NCBISequenceLinkURL
from dark.intervals import ReadIntervals
from dark.features import ProteinFeatureAdder, NucleotideFeatureAdder
from dark import orfs
from dark.intervals import OffsetAdjuster
from dark.score import HigherIsBetterScore


QUERY_COLORS = {
    "A": (1.0, 0.0, 0.0),  # Red.
    "C": (0.0, 0.0, 1.0),  # Blue.
    "G": (0.0, 1.0, 0.0),  # Green.
    "N": (1.0, 0.0, 1.0),  # Purple.
    "T": (1.0, 0.8, 0.0),  # Orange.
    "gap": (0.2, 0.2, 0.2),  # Almost black.
    "match": (0.9, 0.9, 0.9),  # Almost white.
    "*": (0.9, 0.9, 0.9),  # Almost white.
}

DEFAULT_BASE_COLOR = (0.5, 0.5, 0.5)  # Grey

# From http://www.randalolson.com/2014/06/28/how-to-make-beautiful-data-\
# visualizations-in-python-with-matplotlib/
#
# These are the "Tableau 20" colors as RGB.
TABLEAU20: list[tuple[float, float, float]] = [
    (31, 119, 180),
    (174, 199, 232),
    (255, 127, 14),
    (255, 187, 120),
    (44, 160, 44),
    (152, 223, 138),
    (214, 39, 40),
    (255, 152, 150),
    (148, 103, 189),
    (197, 176, 213),
    (140, 86, 75),
    (196, 156, 148),
    (227, 119, 194),
    (247, 182, 210),
    (127, 127, 127),
    (199, 199, 199),
    (188, 189, 34),
    (219, 219, 141),
    (23, 190, 207),
    (158, 218, 229),
]

# Scale the above RGB values to the [0, 1] range, the format matplotlib
# accepts.
for i in range(len(TABLEAU20)):
    r, g, b = TABLEAU20[i]
    TABLEAU20[i] = (r / 255.0, g / 255.0, b / 255.0)

# If we're making a plot that has a log-linear X axis, don't show
# background light grey rectangles for any gap whose (logged) width is less
# than SMALLEST_LOGGED_GAP_TO_DISPLAY.
SMALLEST_LOGGED_GAP_TO_DISPLAY = 20

# Y (Score) axis extra spacing above the best read match. The score of the
# best read HSP will be multiplied by this value, the result converted to
# an int, and then used as the upper bound on the Y axis. Adding 1% seems
# to be a good heuristic.
Y_AXIS_UPPER_PADDING = 1.01

# The default base of the logarithm to use when logLinearXAxis is used to
# produce an alignment graph.
DEFAULT_LOG_LINEAR_X_AXIS_BASE = 1.1


def report(msg):
    print("%s: %s" % (ctime(time()), msg))


def alignmentGraph(
    titlesAlignments,
    title,
    addQueryLines=True,
    showFeatures=True,
    logLinearXAxis=False,
    logBase=DEFAULT_LOG_LINEAR_X_AXIS_BASE,
    rankScores=False,
    colorQueryBases=False,
    createFigure=True,
    showFigure=True,
    readsAx=None,
    imageFile=None,
    quiet=False,
    idList=False,
    xRange="subject",
    showOrfs=True,
):
    """
    Align a set of matching reads against a BLAST or DIAMOND hit.

    @param titlesAlignments: A L{dark.titles.TitlesAlignments} instance.
    @param title: A C{str} sequence title that was matched. We plot the
        reads that hit this title.
    @param addQueryLines: if C{True}, draw query lines in full (these will then
        be partly overdrawn by the HSP match against the subject). These are
        the 'whiskers' that potentially protrude from each side of a query.
    @param showFeatures: if C{True}, look online for features of the subject
        sequence (given by hitId).
    @param logLinearXAxis: if C{True}, convert read offsets so that empty
        regions in the plot we're preparing will only be as wide as their
        logged actual values.
    @param logBase: The base of the logarithm to use if logLinearXAxis is
        C{True}.
    @param: rankScores: If C{True}, change the e-values and bit scores for the
        reads for each title to be their rank (worst to best).
    @param colorQueryBases: if C{True}, color each base of a query string. If
        C{True}, then addQueryLines is meaningless since the whole query is
        shown colored.
    @param createFigure: If C{True}, create a figure and give it a title.
    @param showFigure: If C{True}, show the created figure. Set this to
        C{False} if you're creating a panel of figures or just want to save an
        image (with C{imageFile}).
    @param readsAx: If not None, use this as the subplot for displaying reads.
    @param imageFile: If not None, specifies a filename to write the image to.
    @param quiet: If C{True}, don't print progress / timing output.
    @param idList: a dictionary. The keys is a color and the values is a list
        of read identifiers that should be colored in the respective color.
    @param xRange: set to either 'subject' or 'reads' to indicate the range of
        the X axis.
    @param showOrfs: If C{True}, open reading frames will be displayed.
    """

    startTime = time()

    assert xRange in ("subject", "reads"), 'xRange must be either "subject" or "reads".'

    if createFigure:
        width = 20
        figure = plt.figure(figsize=(width, 20))

    createdReadsAx = readsAx is None

    if showFeatures:
        if showOrfs:
            gs = gridspec.GridSpec(4, 1, height_ratios=[3, 1, 1, 12])
            featureAx = plt.subplot(gs[0, 0])
            orfAx = plt.subplot(gs[1, 0])
            orfReversedAx = plt.subplot(gs[2, 0])
            readsAx = readsAx or plt.subplot(gs[3, 0])
        else:
            gs = gridspec.GridSpec(2, 1, height_ratios=[1, 1])
            featureAx = plt.subplot(gs[0, 0])
            readsAx = readsAx or plt.subplot(gs[1, 0])
    else:
        if showOrfs:
            gs = gridspec.GridSpec(3, 1, height_ratios=[1, 1, 12])
            orfAx = plt.subplot(gs[0, 0])
            orfReversedAx = plt.subplot(gs[1, 0])
            readsAx = readsAx or plt.subplot(gs[2, 0])
        else:
            readsAx = readsAx or plt.subplot(111)

    # Make a deep copy of the title alignments. We're potentially going to
    # change the HSP scores, the X axis offsets, etc., and we don't want to
    # interfere with the data we were passed.
    titleAlignments = deepcopy(titlesAlignments[title])

    readsAlignments = titlesAlignments.readsAlignments
    subjectIsNucleotides = readsAlignments.params.subjectIsNucleotides

    if showOrfs and not subjectIsNucleotides:
        # We cannot show ORFs when displaying protein plots.
        showOrfs = False

    # Allow the class of titlesAlignments to adjust HSPs for plotting,
    # if it has a method for doing so.
    try:
        adjuster = readsAlignments.adjustHspsForPlotting
    except AttributeError:
        pass
    else:
        adjuster(titleAlignments)

    if rankScores:
        reverse = titlesAlignments.scoreClass is not HigherIsBetterScore
        for rank, hsp in enumerate(
            sorted(titleAlignments.hsps(), reverse=reverse), start=1
        ):
            hsp.score.score = rank

    if logLinearXAxis:
        readIntervals = ReadIntervals(titleAlignments.subjectLength)
        # Examine all HSPs so we can build an offset adjuster.
        for hsp in titleAlignments.hsps():
            readIntervals.add(hsp.readStartInSubject, hsp.readEndInSubject)
        # Now adjust offsets in all HSPs.
        offsetAdjuster = OffsetAdjuster(readIntervals, base=logBase)
        for hsp in titleAlignments.hsps():
            offsetAdjuster.adjustHSP(hsp)
        # A function for adjusting other offsets, below.
        adjustOffset = offsetAdjuster.adjustOffset
    else:

        def adjustOffset(offset):
            return offset

    # It would be more efficient to only walk through all HSPs once and
    # compute these values all at once, but for now this is simple and clear.
    maxY = int(ceil(titleAlignments.bestHsp().score.score))
    minY = int(titleAlignments.worstHsp().score.score)
    maxX = max(hsp.readEndInSubject for hsp in titleAlignments.hsps())
    minX = min(hsp.readStartInSubject for hsp in titleAlignments.hsps())

    if xRange == "subject":
        # We'll display a graph for the full subject range. Adjust X axis
        # min/max to make sure we cover at least zero to the sequence length.
        maxX = max(titleAlignments.subjectLength, maxX)
        minX = min(0, minX)

    # Swap min & max Y values, if needed, as it's possible we are dealing
    # with LSPs but that the score adjuster made numerically greater values
    # for those that were small.
    if maxY < minY:
        (maxY, minY) = (minY, maxY)

    if logLinearXAxis:
        # Adjust minX and maxX if we have gaps at the subject start or end.
        gaps = list(readIntervals.walk())
        if gaps:
            # Check start of first gap:
            intervalType, (start, stop) = gaps[0]
            if intervalType == ReadIntervals.EMPTY:
                adjustedStart = adjustOffset(start)
                if adjustedStart < minX:
                    minX = adjustedStart
            # Check stop of last gap:
            intervalType, (start, stop) = gaps[-1]
            if intervalType == ReadIntervals.EMPTY:
                adjustedStop = adjustOffset(stop)
                if adjustedStop > maxX:
                    maxX = adjustedStop

    # We're all set up to start plotting the graph.

    # Add light grey vertical rectangles to show the logarithmic gaps. Add
    # these first so that reads will be plotted on top of them. Only draw
    # gaps that are more than SMALLEST_LOGGED_GAP_TO_DISPLAY pixels wide as
    # we could have millions of tiny gaps for a bacteria and drawing them
    # all will be slow and only serves to make the entire background grey.
    if logLinearXAxis and len(offsetAdjuster.adjustments()) < 100:
        for intervalType, interval in readIntervals.walk():
            if intervalType == ReadIntervals.EMPTY:
                adjustedStart = adjustOffset(interval[0])
                adjustedStop = adjustOffset(interval[1])
                width = adjustedStop - adjustedStart
                if width >= SMALLEST_LOGGED_GAP_TO_DISPLAY:
                    readsAx.axvspan(adjustedStart, adjustedStop, color="#f4f4f4")

    if colorQueryBases:
        # Color each query by its bases.
        xScale = 3
        yScale = 2
        baseImage = BaseImage(
            maxX - minX, maxY - minY + (1 if rankScores else 0), xScale, yScale
        )
        for alignment in titleAlignments:
            for hsp in alignment.hsps:
                y = hsp.score.score - minY
                # If the product of the subject and read frame values is +ve,
                # then they're either both +ve or both -ve, so we just use the
                # read as is. Otherwise, we need to reverse complement it.
                if hsp.subjectFrame * hsp.readFrame > 0:
                    query = alignment.read.sequence
                else:
                    # One of the subject or query has negative sense.
                    query = alignment.read.reverseComplement().sequence
                readStartInSubject = hsp.readStartInSubject
                # There are 3 parts of the query string we need to
                # display. 1) the left part (if any) before the matched
                # part of the subject.  2) the matched part (which can
                # include gaps in the query and/or subject). 3) the right
                # part (if any) after the matched part.  For each part,
                # calculate the ranges in which we have to make the
                # comparison between subject and query.

                # NOTE: never use hsp['origHsp'].gaps to calculate the number
                # of gaps, as this number contains gaps in both subject and
                # query.

                # 1. Left part:
                leftRange = hsp.subjectStart - readStartInSubject

                # 2. Match, middle part:
                middleRange = len(hsp.readMatchedSequence)

                # 3. Right part:
                # Using hsp.readEndInSubject - hsp.subjectEnd to calculate the
                # length of the right part leads to the part being too long.
                # The number of gaps needs to be subtracted to get the right
                # length.
                origQuery = hsp.readMatchedSequence.upper()
                rightRange = (
                    hsp.readEndInSubject - hsp.subjectEnd - origQuery.count("-")
                )

                # 1. Left part.
                xOffset = readStartInSubject - minX
                queryOffset = 0
                for queryIndex in range(leftRange):
                    color = QUERY_COLORS.get(
                        query[queryOffset + queryIndex], DEFAULT_BASE_COLOR
                    )
                    baseImage.set(xOffset + queryIndex, y, color)

                # 2. Match part.
                xOffset = hsp.subjectStart - minX
                xIndex = 0
                queryOffset = hsp.subjectStart - hsp.readStartInSubject
                origSubject = hsp.subjectMatchedSequence
                for matchIndex in range(middleRange):
                    if origSubject[matchIndex] == "-":
                        # A gap in the subject was needed to match the query.
                        # In our graph we keep the subject the same even in the
                        # case where BLAST opened gaps in it, so we compensate
                        # for the gap in the subject by not showing this base
                        # of the query.
                        pass
                    else:
                        if origSubject[matchIndex] == origQuery[matchIndex]:
                            # The query matched the subject at this location.
                            # Matching bases are all colored in the same
                            # 'match' color.
                            color = QUERY_COLORS["match"]
                        else:
                            if origQuery[matchIndex] == "-":
                                # A gap in the query. All query gaps get the
                                # same 'gap' color.
                                color = QUERY_COLORS["gap"]
                            else:
                                # Query doesn't match subject (and is not a
                                # gap).
                                color = QUERY_COLORS.get(
                                    origQuery[matchIndex], DEFAULT_BASE_COLOR
                                )
                        baseImage.set(xOffset + xIndex, y, color)
                        xIndex += 1

                # 3. Right part.
                xOffset = hsp.subjectEnd - minX
                backQuery = query[-rightRange:].upper()
                for queryIndex in range(rightRange):
                    color = QUERY_COLORS.get(backQuery[queryIndex], DEFAULT_BASE_COLOR)
                    baseImage.set(xOffset + queryIndex, y, color)

        readsAx.imshow(
            baseImage.data,
            aspect="auto",
            origin="lower",
            interpolation="nearest",
            extent=[minX, maxX, minY, maxY],
        )
    else:
        # Add horizontal lines for all the query sequences. These will be the
        # grey 'whiskers' in the plots once we (below) draw the matched part
        # on top of part of them.
        if addQueryLines:
            for hsp in titleAlignments.hsps():
                y = hsp.score.score
                line = Line2D(
                    [hsp.readStartInSubject, hsp.readEndInSubject],
                    [y, y],
                    color="#aaaaaa",
                )
                readsAx.add_line(line)

        # Add the horizontal BLAST alignment lines.

        # If an idList is given set things up to look up read colors.
        readColor = {}
        if idList:
            for color, reads in idList.items():
                for read in reads:
                    if read in readColor:
                        raise ValueError(
                            "Read %s is specified multiple times in idList" % read
                        )
                    else:
                        readColor[read] = color

        # Draw the matched region.
        for titleAlignment in titleAlignments:
            readId = titleAlignment.read.id
            for hsp in titleAlignment.hsps:
                y = hsp.score.score
                line = Line2D(
                    [hsp.subjectStart, hsp.subjectEnd],
                    [y, y],
                    color=readColor.get(readId, "blue"),
                )
                readsAx.add_line(line)

    if showOrfs:
        subject = readsAlignments.getSubjectSequence(title)
        orfs.addORFs(orfAx, subject.sequence, minX, maxX, adjustOffset)
        orfs.addReversedORFs(
            orfReversedAx,
            subject.reverseComplement().sequence,
            minX,
            maxX,
            adjustOffset,
        )

    if showFeatures:
        if subjectIsNucleotides:
            featureAdder = NucleotideFeatureAdder()
        else:
            featureAdder = ProteinFeatureAdder()

        features = featureAdder.add(featureAx, title, minX, maxX, adjustOffset)

        # If there are features and there weren't too many of them, add
        # vertical feature lines to the reads and ORF axes.
        if features and not featureAdder.tooManyFeaturesToPlot:
            for feature in features:
                start = feature.start
                end = feature.end
                color = feature.color
                readsAx.axvline(x=start, color=color)
                readsAx.axvline(x=end, color="#cccccc")
                if showOrfs:
                    orfAx.axvline(x=start, color=color)
                    orfAx.axvline(x=end, color="#cccccc")
                    orfReversedAx.axvline(x=start, color=color)
                    orfReversedAx.axvline(x=end, color="#cccccc")
    else:
        features = None

    # We'll return some information we've gathered.
    result = {
        "adjustOffset": adjustOffset,
        "features": features,
        "minX": minX,
        "maxX": maxX,
        "minY": minY,
        "maxY": maxY,
    }

    # Allow the class of titlesAlignments to add to the plot, if it has a
    # method for doing so.
    try:
        adjuster = readsAlignments.adjustPlot
    except AttributeError:
        pass
    else:
        adjuster(readsAx)

    # Titles, axis, etc.
    if createFigure:
        readCount = titleAlignments.readCount()
        hspCount = titleAlignments.hspCount()
        figure.suptitle(
            "%s\nLength %d %s, %d read%s, %d HSP%s."
            % (
                fill(titleAlignments.subjectTitle, 80),
                titleAlignments.subjectLength,
                "nt" if subjectIsNucleotides else "aa",
                readCount,
                "" if readCount == 1 else "s",
                hspCount,
                "" if hspCount == 1 else "s",
            ),
            fontsize=20,
        )

    # Add a title and y-axis label, but only if we made the reads axes.
    if createdReadsAx:
        readsAx.set_title("Read alignments", fontsize=20)
        ylabel = readsAlignments.params.scoreTitle
        if rankScores:
            ylabel += " rank"
        plt.ylabel(ylabel, fontsize=17)

    # Set the x-axis limits.
    readsAx.set_xlim([minX - 1, maxX + 1])

    readsAx.set_ylim([0, int(maxY * Y_AXIS_UPPER_PADDING)])
    readsAx.grid()
    if createFigure:
        if showFigure:
            plt.show()
        if imageFile:
            figure.savefig(imageFile)
    stop = time()
    if not quiet:
        report("Graph generated in %.3f mins." % ((stop - startTime) / 60.0))

    return result


def alignmentPanel(
    titlesAlignments,
    sortOn="maxScore",
    idList=False,
    equalizeXAxes=False,
    xRange="subject",
    logLinearXAxis=False,
    rankScores=False,
    showFeatures=True,
    logBase=DEFAULT_LOG_LINEAR_X_AXIS_BASE,
):
    """
    Produces a rectangular panel of graphs that each contain an alignment graph
    against a given sequence.

    @param titlesAlignments: A L{dark.titles.TitlesAlignments} instance.
    @param sortOn: The attribute to sort subplots on. Either "maxScore",
        "medianScore", "readCount", "length", or "title".
    @param idList: A dictionary. Keys are colors and values are lists of read
        ids that should be colored using that color.
    @param equalizeXAxes: If C{True}, adjust the X axis on each alignment plot
        to be the same.
    @param xRange: Set to either 'subject' or 'reads' to indicate the range of
        the X axis.
    @param logLinearXAxis: If C{True}, convert read offsets so that empty
        regions in the plots we're preparing will only be as wide as their
        logged actual values.
    @param logBase: The logarithm base to use if logLinearXAxis is C{True}.
    @param: rankScores: If C{True}, change the scores for the reads for each
        title to be their rank (worst to best).
    @param showFeatures: If C{True}, look online for features of the subject
        sequences.
    @raise ValueError: If C{outputDir} exists but is not a directory or if
        C{xRange} is not "subject" or "reads".
    """

    if xRange not in ("subject", "reads"):
        raise ValueError('xRange must be either "subject" or "reads".')

    start = time()
    titles = titlesAlignments.sortTitles(sortOn)
    cols = 5
    rows = int(len(titles) / cols) + (0 if len(titles) % cols == 0 else 1)
    figure, ax = plt.subplots(rows, cols, squeeze=False)
    allGraphInfo = {}
    coords = dimensionalIterator((rows, cols))

    report(
        "Plotting %d titles in %dx%d grid, sorted on %s"
        % (len(titles), rows, cols, sortOn)
    )

    for i, title in enumerate(titles):
        titleAlignments = titlesAlignments[title]
        row, col = next(coords)
        report("%d: %s %s" % (i, title, NCBISequenceLinkURL(title, "")))

        # Add a small plot to the alignment panel.
        graphInfo = alignmentGraph(
            titlesAlignments,
            title,
            addQueryLines=True,
            showFeatures=showFeatures,
            rankScores=rankScores,
            logLinearXAxis=logLinearXAxis,
            logBase=logBase,
            colorQueryBases=False,
            createFigure=False,
            showFigure=False,
            readsAx=ax[row][col],
            quiet=True,
            idList=idList,
            xRange=xRange,
            showOrfs=False,
        )

        allGraphInfo[title] = graphInfo
        readCount = titleAlignments.readCount()
        hspCount = titleAlignments.hspCount()

        # Make a short title for the small panel blue plot, ignoring any
        # leading NCBI gi / accession numbers.
        if title.startswith("gi|") and title.find(" ") > -1:
            shortTitle = title.split(" ", 1)[1][:40]
        else:
            shortTitle = title[:40]

        plotTitle = "%d: %s\nLength %d, %d read%s, %d HSP%s." % (
            i,
            shortTitle,
            titleAlignments.subjectLength,
            readCount,
            "" if readCount == 1 else "s",
            hspCount,
            "" if hspCount == 1 else "s",
        )

        if hspCount:
            if rankScores:
                plotTitle += "\nY axis is ranked score"
            else:
                plotTitle += "\nmax %.2f, median %.2f" % (
                    titleAlignments.bestHsp().score.score,
                    titleAlignments.medianScore(),
                )

        ax[row][col].set_title(plotTitle, fontsize=10)

    maxX = max(graphInfo["maxX"] for graphInfo in allGraphInfo.values())
    minX = min(graphInfo["minX"] for graphInfo in allGraphInfo.values())
    maxY = max(graphInfo["maxY"] for graphInfo in allGraphInfo.values())
    minY = min(graphInfo["minY"] for graphInfo in allGraphInfo.values())

    # Post-process graphs to adjust axes, etc.

    coords = dimensionalIterator((rows, cols))
    for title in titles:
        titleAlignments = titlesAlignments[title]
        row, col = next(coords)
        a = ax[row][col]
        a.set_ylim([0, int(maxY * Y_AXIS_UPPER_PADDING)])
        if equalizeXAxes:
            a.set_xlim([minX, maxX])
        a.set_yticks([])
        a.set_xticks([])

        if xRange == "subject" and minX < 0:
            # Add a vertical line at x=0 so we can see the 'whiskers' of
            # reads that extend to the left of the sequence we're aligning
            # against.
            a.axvline(x=0, color="#cccccc")

        # Add a line on the right of each sub-plot so we can see where the
        # sequence ends (as all panel graphs have the same width and we
        # otherwise couldn't tell).
        sequenceLen = titleAlignments.subjectLength
        if logLinearXAxis:
            sequenceLen = allGraphInfo[title]["adjustOffset"](sequenceLen)
        a.axvline(x=sequenceLen, color="#cccccc")

    # Hide the final panel graphs (if any) that have no content. We do this
    # because the panel is a rectangular grid and some of the plots at the
    # end of the last row may be unused.
    for row, col in coords:
        ax[row][col].axis("off")

    # plt.subplots_adjust(left=0.01, bottom=0.01, right=0.99, top=0.93,
    # wspace=0.1, hspace=None)
    plt.subplots_adjust(hspace=0.4)
    figure.suptitle(
        "X: %d to %d, Y (%s): %d to %d"
        % (
            minX,
            maxX,
            titlesAlignments.readsAlignments.params.scoreTitle,
            int(minY),
            int(maxY),
        ),
        fontsize=20,
    )
    figure.set_size_inches(5 * cols, 3 * rows, forward=True)
    figure.show()
    stop = time()
    report("Alignment panel generated in %.3f mins." % ((stop - start) / 60.0))


def alignmentPanelHTML(
    titlesAlignments,
    sortOn="maxScore",
    outputDir=None,
    idList=False,
    equalizeXAxes=False,
    xRange="subject",
    logLinearXAxis=False,
    logBase=DEFAULT_LOG_LINEAR_X_AXIS_BASE,
    rankScores=False,
    showFeatures=True,
    showOrfs=True,
):
    """
    Produces an HTML index file in C{outputDir} and a collection of alignment
    graphs and FASTA files to summarize the information in C{titlesAlignments}.

    @param titlesAlignments: A L{dark.titles.TitlesAlignments} instance.
    @param sortOn: The attribute to sort subplots on. Either "maxScore",
        "medianScore", "readCount", "length", or "title".
    @param outputDir: Specifies a C{str} directory to write the HTML to. If
        the directory does not exist it will be created.
    @param idList: A dictionary. Keys are colors and values are lists of read
        ids that should be colored using that color.
    @param equalizeXAxes: If C{True}, adjust the X axis on each alignment plot
        to be the same.
    @param xRange: Set to either 'subject' or 'reads' to indicate the range of
        the X axis.
    @param logLinearXAxis: If C{True}, convert read offsets so that empty
        regions in the plots we're preparing will only be as wide as their
        logged actual values.
    @param logBase: The logarithm base to use if logLinearXAxis is C{True}.
    @param: rankScores: If C{True}, change the scores for the reads for each
        title to be their rank (worst to best).
    @param showFeatures: If C{True}, look online for features of the subject
        sequences.
    @param showOrfs: If C{True}, open reading frames will be displayed.
    @raise TypeError: If C{outputDir} is C{None}.
    @raise ValueError: If C{outputDir} is None or exists but is not a
        directory or if C{xRange} is not "subject" or "reads".
    """

    if xRange not in ("subject", "reads"):
        raise ValueError('xRange must be either "subject" or "reads".')

    if equalizeXAxes:
        raise NotImplementedError("This feature is not yet implemented.")

    titles = titlesAlignments.sortTitles(sortOn)

    if os.access(outputDir, os.F_OK):
        # outputDir exists. Check it's a directory.
        if not S_ISDIR(os.stat(outputDir).st_mode):
            raise ValueError("%r is not a directory." % outputDir)
    else:
        if outputDir is None:
            raise ValueError("The outputDir needs to be specified.")
        else:
            os.mkdir(outputDir)

    htmlWriter = AlignmentPanelHTMLWriter(outputDir, titlesAlignments)

    for i, title in enumerate(titles):
        # titleAlignments = titlesAlignments[title]

        # If we are writing data to a file too, create a separate file with
        # a plot (this will be linked from the summary HTML).
        imageBasename = "%d.png" % i
        imageFile = "%s/%s" % (outputDir, imageBasename)
        graphInfo = alignmentGraph(
            titlesAlignments,
            title,
            addQueryLines=True,
            showFeatures=showFeatures,
            rankScores=rankScores,
            logLinearXAxis=logLinearXAxis,
            logBase=logBase,
            colorQueryBases=False,
            showFigure=False,
            imageFile=imageFile,
            quiet=True,
            idList=idList,
            xRange=xRange,
            showOrfs=showOrfs,
        )

        # Close the image plot to make sure memory is flushed.
        plt.close()
        htmlWriter.addImage(imageBasename, title, graphInfo)

    htmlWriter.close()


def scoreGraph(
    titlesAlignments, find=None, showTitles=False, figureWidth=5, figureHeight=5
):
    """
    NOTE: This function has probably bit rotted (but only a little).

    Produce a rectangular panel of graphs, each of which shows sorted scores
    for a title. Matches against a certain sequence title, as determined by
    C{find}, (see below) are highlighted.

    @param find: A function that can be passed a sequence title. If the
        function returns C{True} a red dot is put into the graph at that point
        to highlight the match.
    @param showTitles: If C{True} display read sequence names. The panel tends
        to look terrible when titles are displayed. If C{False}, show no title.
    @param figureWidth: The C{float} width of the figure, in inches.
    @param figureHeight: The C{float} height of the figure, in inches.
    """
    maxScore = None
    maxHsps = 0
    cols = 5
    rows = int(len(titlesAlignments) / cols) + (
        0 if len(titlesAlignments) % cols == 0 else 1
    )
    f, ax = plt.subplots(rows, cols)
    coords = dimensionalIterator((rows, cols))

    for title in titlesAlignments:
        titleAlignments = titlesAlignments[title]
        row, col = next(coords)
        hspCount = titleAlignments.hspCount()
        if hspCount > maxHsps:
            maxHsps = hspCount
        scores = []
        highlightX = []
        highlightY = []
        for x, titleAlignment in enumerate(titleAlignments):
            score = titleAlignment.hsps[0].score.score
            scores.append(score)
            if find and find(titleAlignment.subjectTitle):
                highlightX.append(x)
                highlightY.append(score)
        a = ax[row][col]
        if scores:
            max_ = max(scores)
            if maxScore is None or max_ > maxScore:
                maxScore = max_
            x = np.arange(0, len(scores))
            a.plot(x, scores)
        if highlightX:
            a.plot(highlightX, highlightY, "ro")
        if showTitles:
            a.set_title("%s" % title, fontsize=10)

    # Adjust all plots to have the same dimensions.
    coords = dimensionalIterator((rows, cols))
    for _ in range(len(titlesAlignments)):
        row, col = next(coords)
        a = ax[row][col]
        a.axis([0, maxHsps, 0, maxScore])
        # a.set_yscale('log')
        a.set_yticks([])
        a.set_xticks([])

    # Hide the final panel graphs (if any) that have no content. We do this
    # because the panel is a rectangular grid and some of the plots at the
    # end of the last row may be unused.
    for row, col in coords:
        ax[row][col].axis("off")

    plt.subplots_adjust(
        left=0.01, bottom=0.01, right=0.99, top=0.93, wspace=0.1, hspace=None
    )
    f.suptitle("max HSPs %d, max score %f" % (maxHsps, maxScore))
    f.set_size_inches(figureWidth, figureHeight, forward=True)
    # f.savefig('scores.png')
    plt.show()


def scatterAlign(seq1, seq2, window=7):
    """
    Visually align two sequences.
    """
    d1 = defaultdict(list)
    d2 = defaultdict(list)
    for seq, section_dict in [(seq1, d1), (seq2, d2)]:
        for i in range(len(seq) - window):
            section = seq[i : i + window]
            section_dict[section].append(i)
    matches = set(d1).intersection(d2)
    print("%i unique matches" % len(matches))
    x = []
    y = []
    for section in matches:
        for i in d1[section]:
            for j in d2[section]:
                x.append(i)
                y.append(j)
    # plt.cla()  # clear any prior graph
    plt.gray()
    plt.scatter(x, y)
    plt.xlim(0, len(seq1) - window)
    plt.ylim(0, len(seq2) - window)
    plt.xlabel("length %i bp" % (len(seq1)))
    plt.ylabel("length %i bp" % (len(seq2)))
    plt.title("Dot plot using window size %i\n(allowing no mis-matches)" % window)
    plt.show()


def plotAAProperties(sequence, propertyNames, showLines=True, showFigure=True):
    """
    Plot amino acid property values for a sequence.

    @param sequence: An C{AARead} (or a subclass) instance.
    @param propertyNames: An iterable of C{str} property names (each of which
        must be a key of a key in the C{dark.aa.PROPERTY_DETAILS} C{dict}).
    @param showLines: If C{True}, lines will be drawn between successive AA
        property values. If not, just the values will be plotted as a scatter
        plot (this greatly reduces visual clutter if the sequence is long and
        AA property values are variable).
    @param showFigure: If C{True}, display the plot. Passing C{False} is useful
        in testing.
    @raise ValueError: If an unknown property is given in C{propertyNames}.
    @return: The return value from calling dark.aa.propertiesForSequence:
        a C{dict} keyed by (lowercase) property name, with values that are
        C{list}s of the corresponding property value according to sequence
        position.
    """
    MISSING_AA_VALUE = -1.1
    propertyValues = propertiesForSequence(
        sequence, propertyNames, missingAAValue=MISSING_AA_VALUE
    )
    if showFigure:
        legend = []
        x = np.arange(0, len(sequence))
        plot = plt.plot if showLines else plt.scatter

        for index, propertyName in enumerate(propertyValues):
            color = TABLEAU20[index]
            plot(x, propertyValues[propertyName], color=color)
            legend.append(patches.Patch(color=color, label=propertyName))

        plt.legend(handles=legend, loc=(0, 1.1))
        plt.xlim(-0.2, len(sequence) - 0.8)
        plt.ylim(min(MISSING_AA_VALUE, -1.1), 1.1)
        plt.xlabel("Sequence index")
        plt.ylabel("Property value")
        plt.title(sequence.id)
        plt.show()

    return propertyValues


def plotAAClusters(sequence, propertyNames, showLines=True, showFigure=True):
    """
    Plot amino acid property cluster numbers for a sequence.

    @param sequence: An C{AARead} (or a subclass) instance.
    @param propertyNames: An iterable of C{str} property names (each of which
        must be a key of a key in the C{dark.aa.PROPERTY_CLUSTERS} C{dict}).
    @param showLines: If C{True}, lines will be drawn between successive AA
        property values. If not, just the values will be plotted as a scatter
        plot (this greatly reduces visual clutter if the sequence is long and
        AA property values are variable).
    @param showFigure: If C{True}, display the plot. Passing C{False} is useful
        in testing.
    @raise ValueError: If an unknown property is given in C{propertyNames}.
    @return: The return value from calling dark.aa.clustersForSequence:
        a C{dict} keyed by (lowercase) property name, with values that are
        C{list}s of the corresponding property value according to sequence
        position.
    """
    MISSING_AA_VALUE = 0
    propertyClusters = clustersForSequence(
        sequence, propertyNames, missingAAValue=MISSING_AA_VALUE
    )
    if showFigure:
        minCluster = 1
        maxCluster = -1
        legend = []
        x = np.arange(0, len(sequence))
        plot = plt.plot if showLines else plt.scatter

        for index, propertyName in enumerate(propertyClusters):
            color = TABLEAU20[index]
            clusterNumbers = propertyClusters[propertyName]
            plot(x, clusterNumbers, color=color)
            legend.append(patches.Patch(color=color, label=propertyName))

            propertyMinCluster = min(clusterNumbers)
            if propertyMinCluster < minCluster:
                minCluster = propertyMinCluster

            propertyMaxCluster = max(clusterNumbers)
            if propertyMaxCluster > maxCluster:
                maxCluster = propertyMaxCluster

        plt.legend(handles=legend, loc=(0, 1.1))
        plt.xlim(-0.2, len(sequence) - 0.8)
        plt.ylim(minCluster - 0.5, maxCluster + 0.5)
        plt.yticks(range(maxCluster + 1))
        plt.xlabel("Sequence index")
        plt.ylabel("Property cluster number")
        plt.title(sequence.id)
        plt.show()

    return propertyClusters



#!/usr/bin/env python

from Bio.Seq import translate

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Translate FASTA and write the result to standard output."
    )

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()

    for read in parseFASTACommandLineOptions(args):
        print(f">{read.id}\n{translate(read.sequence)}")



#!/usr/bin/env python

import sys
import argparse
from math import log10
import multiprocessing

from dark.aligners import edlibAlign, mafft, needle
from dark.dna import compareDNAReads, matchToString, AMBIGUOUS
from dark.reads import Reads, addFASTACommandLineOptions, parseFASTACommandLineOptions
from dark.utils import parseRangeExpression

MAFFT_DEFAULT_ARGS = "--globalpair --maxiterate 1000 --preservecase"
MAFFT_ALGORITHMS_URL = (
    "https://mafft.cbrc.jp/alignment/software/algorithms/algorithms.html"
)
NEEDLE_DEFAULT_ARGS = "auto"


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description="Compare two sequences.",
)

parser.add_argument(
    "--index1",
    type=int,
    default=1,
    help="The (1-based) index in the input of the first sequence.",
)

parser.add_argument(
    "--index2",
    type=int,
    default=2,
    help="The (1-based) index in the input of the second sequence.",
)

parser.add_argument(
    "--align",
    action="store_true",
    help=(
        "If given, use mafft (the default) or needle (according to the "
        "algorithm selected by --aligner) to align the two sequences."
    ),
)

parser.add_argument(
    "--aligner",
    default="mafft",
    choices=("edlib", "mafft", "needle"),
    help="The alignment algorithm to use.",
)

parser.add_argument(
    "--alignerOptions",
    help=(
        "Optional arguments to pass to the alignment algorithm. If the "
        'aligner is mafft, the default options are %r. If needle, "%s". '
        "Do not try to set the number of threads here - use the --threads "
        "argument instead. If you are using mafft, see %s for some possible "
        "option combinations."
        % (MAFFT_DEFAULT_ARGS, NEEDLE_DEFAULT_ARGS, MAFFT_ALGORITHMS_URL)
    ),
)

parser.add_argument(
    "--threads",
    type=int,
    default=multiprocessing.cpu_count(),
    help=(
        "The number of threads to use when running the aligner (if --align "
        "is used and the alignment algorithm can make use of multiple "
        "threads (mafft can, needle cannot))."
    ),
)

parser.add_argument(
    "--alignmentFile", help="The file to save the alignment to (implies --align)."
)

parser.add_argument(
    "--strict",
    action="store_true",
    help="If given, do not allow ambiguous nucleotide symbols to match.",
)

parser.add_argument(
    "--quiet",
    dest="verbose",
    action="store_false",
    help=("Do not print information about aligning, or falling back to " "stretcher."),
)

parser.add_argument(
    "--noGapLocations",
    dest="includeGapLocations",
    action="store_false",
    help="Do not indicate the (1-based) locations of sequence gaps.",
)

parser.add_argument(
    "--noNoCoverageLocations",
    dest="includeCoverageLocations",
    action="store_false",
    help="Do not indicate the (1-based) locations of no coverage.",
)

parser.add_argument(
    "--sites",
    help=(
        "Specify (1-based) sequence sites to keep. All other sites will "
        "be ignored. The sites must be given in the form e.g., "
        "24,100-200,260."
    ),
)

parser.add_argument(
    "--showDiffs",
    action="store_true",
    help="Print (1-based) sites where the sequence nucleotides differ.",
)

parser.add_argument(
    "--showAmbiguous",
    action="store_true",
    help=(
        "Print (1-based) sites where either sequence has an ambiguous "
        "nucleotide code."
    ),
)

parser.add_argument(
    "--includeAmbiguousMatches",
    action="store_true",
    help=(
        "Print (1-based) sites of ambiguous matches. The output gives the "
        "site, the base from the first sequence, then the base from the "
        "second sequence."
    ),
)

parser.add_argument(
    "--includeNonGapMismatches",
    action="store_true",
    help=(
        "Print (1-based) sites of mismatches that do not involve a gap. The "
        "output gives the site, the base from the first sequence, then the "
        "base from the second sequence."
    ),
)

parser.add_argument(
    "--gapChars",
    default="-",
    metavar="CHARS",
    help=(
        "The sequence characters that should be considered to be gaps. "
        "These characters will be ignored in computing sequence lengths "
        "and identity fractions."
    ),
)

parser.add_argument(
    "--noCoverageChars",
    metavar="CHARS",
    help=(
        "The sequence characters that indicate lack of coverage. "
        "These characters will be ignored in identity fractions."
    ),
)

addFASTACommandLineOptions(parser)
args = parser.parse_args()

keepSequences = set([args.index1 - 1, args.index2 - 1])

reads = list(parseFASTACommandLineOptions(args).filter(keepSequences=keepSequences))

if len(reads) == 1:
    if len(keepSequences) == 1:
        # This is ok, they want to compare a sequence with itself.
        reads = Reads([reads[0], reads[0]])
    else:
        print(
            "Could not find both requested sequence indices. Exiting.", file=sys.stderr
        )
        sys.exit(1)
elif len(reads) != 2:
    print("Could not find both requested sequence indices. Exiting.", file=sys.stderr)
    sys.exit(1)

if args.alignmentFile:
    args.align = True

if args.align:
    len1, len2 = map(len, reads)
    if len1 == len2:
        print("Pre-alignment, sequence lengths were identical: %d" % len1)
    else:
        print(
            "Pre-alignment, sequence lengths: %d, %d (difference %d)"
            % (len1, len2, abs(len1 - len2))
        )

    print("  Gaps:")
    print("    Id: %s %d" % (reads[0].id, reads[0].sequence.count("-")))
    print("    Id: %s %d" % (reads[1].id, reads[1].sequence.count("-")))

    if args.aligner == "mafft":
        # Be careful in examining args.alignerOptions because we want the
        # user to be able to pass an empty string (so check against None
        # before deciding to use the default.)
        options = (
            MAFFT_DEFAULT_ARGS if args.alignerOptions is None else args.alignerOptions
        )
        reads = mafft(reads, args.verbose, options=options, threads=args.threads)
    elif args.aligner == "needle":
        # Be careful in examining args.alignerOptions because we want the
        # user to be able to pass an empty string (so check against None
        # before deciding to use the default.)
        options = (
            NEEDLE_DEFAULT_ARGS if args.alignerOptions is None else args.alignerOptions
        )
        reads = needle(reads, args.verbose, options=options)
    else:
        assert args.aligner == "edlib"
        reads = edlibAlign(reads)

    if args.alignmentFile:
        assert reads.save(args.alignmentFile) == 2

offsets = (
    parseRangeExpression(args.sites, convertToZeroBased=True) if args.sites else None
)

read1, read2 = reads
len1, len2 = map(len, reads)
identicalLengths = len1 == len2

# Sanity check.
if args.align:
    assert identicalLengths

match = compareDNAReads(
    read1,
    read2,
    matchAmbiguous=(not args.strict),
    offsets=offsets,
    gapChars=args.gapChars,
    noCoverageChars=args.noCoverageChars,
)

x = "Post-alignment, sequence" if args.align else "Sequence"
if identicalLengths:
    print("%s lengths are identical: %s" % (x, len1))
else:
    print("%s lengths: %d, %d (difference %d)" % (x, len1, len2, abs(len1 - len2)))

print(
    matchToString(
        match,
        read1,
        read2,
        matchAmbiguous=(not args.strict),
        offsets=offsets,
        includeGapLocations=args.includeGapLocations,
        includeNoCoverageLocations=args.includeCoverageLocations,
        includeAmbiguousMatches=args.includeAmbiguousMatches,
        includeNonGapMismatches=args.includeNonGapMismatches,
    )
)

if args.showDiffs:
    # Print all sites where the sequences differ.
    width = int(log10(max(len1, len2))) + 1
    headerPrinted = False
    for site, (a, b) in enumerate(zip(read1.sequence, read2.sequence), start=1):
        if a != b:
            if not headerPrinted:
                print("Differences (site, %s, %s):" % (read1.id, read2.id))
                headerPrinted = True
            print("  %*d %s %s" % (width, site, a, b))

    if not headerPrinted:
        print("No sequence differences found.")

if args.showAmbiguous:
    width = int(log10(max(len1, len2))) + 1
    headerPrinted = False
    for site, (a, b) in enumerate(zip(read1.sequence, read2.sequence), start=1):
        if len(AMBIGUOUS.get(a, "")) > 1 or len(AMBIGUOUS.get(b, "")) > 1:
            if not headerPrinted:
                print("Ambiguities (site, %s, %s):" % (read1.id, read2.id))
                headerPrinted = True
            print("  %*d %s %s" % (width, site, a, b))

    if not headerPrinted:
        print("No sequence ambiguities found.")



from os import environ

from mysql import connector


def getDatabaseConnection():
    return connector.connect(
        host="localhost",
        user=environ.get("DBI_USER", environ["USER"]),
        password=environ["DBI_PASSWORD"],
        database="ncbi_taxonomy",
        buffered=True,
    )



#!/usr/bin/env python

import sys
import argparse
from numpy import std
from collections import defaultdict

from dark.filter import (
    addFASTAFilteringCommandLineOptions,
    parseFASTAFilteringCommandLineOptions,
)
from dark.reads import Reads
from dark.sam import samfile, SAMFilter, samReferences, UnknownReference
from dark.utils import baseCountsToStr, pct


def makeParser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Print SAM/BAM file coverage statistics by offset. "
            "Output lines show the offset."
        ),
    )

    addFASTAFilteringCommandLineOptions(parser)
    SAMFilter.addFilteringOptions(parser, samfileIsPositional=True)

    parser.add_argument(
        "--noOffsets",
        action="store_false",
        dest="printOffsets",
        help="Do not print per-offset details of base counts.",
    )

    parser.add_argument(
        "--noStats",
        action="store_false",
        dest="printStats",
        help="Do not print final average and standard deviation statistics.",
    )

    parser.add_argument(
        "--noFilter",
        action="store_false",
        dest="filter",
        help=(
            "Do not use our SAM filtering. Note that if you give this option, "
            "any filtering option (other than --referenceId) you also specify "
            "that is provided by the SAMFilter.addFilteringOptions will be "
            "silently ignored!"
        ),
    )

    return parser


def main():
    parser = makeParser()
    args = parser.parse_args()

    if not (args.printOffsets or args.printStats):
        print(
            "You have used both --noOffsets and --noStats, so there is no output!",
            file=sys.stderr,
        )
        sys.exit(1)

    # We don't have a file of reads, we just want a read filter that we can use
    # to filter the SAM file query sequences and to get reference lengths from.
    reads = parseFASTAFilteringCommandLineOptions(args, Reads())
    samFilter = SAMFilter.parseFilteringOptions(args, reads.filterRead)

    if samFilter.referenceIds and len(samFilter.referenceIds) > 1:
        print(
            "Only one reference id can be given. To calculate coverage for more "
            "than one reference, run this script multiple times.",
            file=sys.stderr,
        )
        sys.exit(1)

    try:
        referenceLengths = samFilter.referenceLengths()
    except UnknownReference:
        referenceId = samFilter.referenceIds.pop()
        referenceIds = samReferences(args.samfile)
        print(
            "Reference %r does not appear in SAM file %s. Known "
            "references are: %s."
            % (referenceId, args.samfile, ", ".join(sorted(referenceIds))),
            file=sys.stderr,
        )
        sys.exit(1)

    if args.filter:

        def filterRead(read):
            return not (read.is_del or read.is_refskip) and samFilter.filterAlignment(
                read.alignment
            )

    else:
        # Do not do our custom SAM filtering.
        def filterRead(read):
            return not (read.is_del or read.is_refskip)

    if args.printStats:
        counts = []

    with samfile(args.samfile) as sam:
        if samFilter.referenceIds:
            # No need to check if the given reference id is in referenceLengths
            # because the samFilter.referenceLengths call above catches that.
            referenceId = samFilter.referenceIds.pop()
        else:
            if len(referenceLengths) == 1:
                referenceId = list(referenceLengths)[0]
            else:
                print(
                    "SAM file %r contains %d references (%s). Only one "
                    "reference id can be analyzed at a time. Please use "
                    "--referenceId to specify the one you want examined."
                    % (
                        args.samfile,
                        len(referenceLengths),
                        ", ".join(sorted(referenceLengths)),
                    ),
                    file=sys.stderr,
                )
                sys.exit(1)

        for column in sam.pileup(
            reference=referenceId,
            start=0,
            stop=referenceLengths[referenceId],
            truncate=True,
        ):
            bases = defaultdict(int)
            for read in column.pileups:
                if filterRead(read):
                    base = read.alignment.query_sequence[read.query_position]
                    bases[base] += 1

            baseCount = sum(bases.values())

            if args.printStats:
                counts.append(baseCount)

            if args.printOffsets:
                print(
                    f"{column.reference_pos + 1} {baseCount} "
                    f"{baseCountsToStr(bases)}"
                )

    if args.printStats:
        referenceLength = referenceLengths[referenceId]
        print("Reference id: %s" % referenceId)
        print("Reference length: %d" % referenceLength)
        print("Bases covered: %s" % pct(len(counts), referenceLength))
        print(
            "Min coverage depth: %d"
            % (0 if len(counts) < referenceLength else min(counts))
        )
        print("Max coverage depth: %d" % max(counts, default=0))
        print("Mean coverage depth: %.3f" % (sum(counts) / referenceLength))
        if counts:
            print("Coverage depth s.d.: %.3f" % std(counts))


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import sys
import argparse
from pathlib import Path
import dendropy
from warnings import warn


def parseArgs():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Curate phylogenetic trees (i.e., their specification files).",
    )

    parser.add_argument(
        "--treeFiles",
        nargs="+",
        help="The tree files to curate.",
    )

    parser.add_argument(
        "--outSuffix",
        default=".treefile-curated",
        help="The suffix to use on curated output files.",
    )

    parser.add_argument(
        "--dryRun",
        action="store_true",
        help="Just print what would be done.",
    )

    parser.add_argument(
        "--force",
        action="store_true",
        help="Overwrite pre-existing result files.",
    )

    parser.add_argument(
        "--rootNode",
        help="Re-root the tree on the incoming branch to the tip node with this name.",
    )

    parser.add_argument(
        "--collapse",
        type=float,
        help=(
            "A floating-point support value. Branches with support less than this "
            "value will be collapsed (to form polytomies). The value to use will depend "
            "on what the program used to create the tree used to indicate the support. "
            "This is frequently the number of times the branch was supported in the "
            "bootstrap analysis. In this case, if you did 100 bootstraps, you would see "
            "integer support values from 0 to 100. If you did 1000 bootstraps, the "
            "values could go up to 1000. Other programs might write the support value "
            "as a fraction (of bootstraps in which the branch is supported), therefore "
            "ranging from 0.0 to 1.0. In other words, you need to know what's in your "
            "tree file to know for sure what value to use for this option. "
            "More difficult is to know what cut-off value actually makes sense in your "
            "analysis. If you're using iqtree2 with its ultrafast bootstrap (UFBoot), "
            "you should read the related FAQ item at "
            "http://www.iqtree.org/doc/Frequently-Asked-Questions"
        ),
    )

    parser.add_argument(
        "--discardLengths",
        action="store_true",
        help=(
            "When collapsing edges due to low support, the length of the edge being "
            "removed will be added to the length of the edges leading to the nodes of "
            "the edge-to-be-removed's child edges. This may be hard to follow, in "
            "which case you should draw yourself a little picture of part of a rooted "
            "tree and think about what happens to the nodes that become detached when "
            "you remove an edge (and the node that it leads to in the direction away "
            "from the root). If the length of the edge you are collapsing is "
            "important, because it came from a tree-making algorithm that sets branch "
            "lengths according to sequence change, the 'right' thing to do is to add "
            "the length of the to-be-deleted edge to the lengths of its descendant "
            "edges when you attach those nodes as children of the node at the root-end "
            "of the edge you are deleting. If for some reason you also want to discard "
            "the length of the edge you are collapsing, you can use this "
            "--discardLengths option. Make sure you really understand the implications "
            "of doing so, however, because discarding that length will cause the "
            "patristic distance of these child nodes to be shorter than it was "
            "estimated to be by the algorithm that made the tree and set the original "
            "branch lengths. And why would you want to second-guess the algorithm and "
            "mess with its conclusions?"
        ),
    )

    parser.add_argument(
        "--ladderize",
        choices=("ascending", "descending"),
        help="How to (optionally) ladderize the tree after re-rooting.",
    )

    parser.add_argument(
        "--format",
        default="newick",
        choices=("fasta", "phylip", "newick", "nexml", "nexus"),
        help="The output format.",
    )

    parser.add_argument(
        "--scale",
        default=1.0,
        type=float,
        help=(
            "Factor to scale the length of the branches leading to the descendants "
            "of the new root (if --rootNode is used). This can be used to reduce "
            "the length of very long branches that result from rerooting on a distant "
            "outgroup. If you use this option for an image in a publication, you of "
            "course must point out that you have done so!"
        ),
    )

    return parser.parse_args()


def main():
    args = parseArgs()

    if not (args.rootNode or args.ladderize or args.collapse):
        exit(
            "Nothing to do! Use some combination of --rootNode (for re-rooting), "
            "--ladderize, or --collapse if you want something done. Exiting."
        )

    for treefile in map(Path, args.treeFiles):
        curated = treefile.parent / f"{treefile.stem}{args.outSuffix}"
        if curated.exists() and not args.force:
            print(
                f"Output {str(curated)!r} exists. Skipping. Use --force to overwrite.",
                file=sys.stderr,
            )
            continue

        with open(treefile) as fp:
            tree = dendropy.Tree.get(
                file=fp, schema="newick", preserve_underscores=True
            )

        # Nodes must either have a taxon (for tips) or a 'label' (interpreted as the
        # support value on the incoming branch of internal nodes).
        for node in tree.preorder_node_iter():
            if node.label and node.taxon:
                # It's sufficiently weird to find a label on a tip branch that we should
                # wonder a) whether the code that created the support values was sane,
                # and b) whether these labels are in fact support values. If the value
                # is 100.0 there's no harm in ignoring it, but we issue a warning. If
                # we've been given a tree with branch labels that are numeric but not
                # support values, a taxon node with a label on it is a pretty good sign
                # that something is amiss.
                msg = (
                    f"Node {node!r} has a label and a taxon. It should have one or "
                    "the other, not both. The label is supposed to be a support "
                    "value, and tips (taxa) normally won't have one because they "
                    "always have 100% support."
                )
                if float(node.label) == 100.0:
                    warn(msg)
                else:
                    exit(msg)

        if args.collapse is not None:
            for node in tree.postorder_node_iter():
                if (
                    node.is_internal()
                    and node.label is not None
                    and float(node.label) < args.collapse
                ):
                    # Collapse this node's incoming edge so its children become children
                    # of this node's parent.
                    node.edge.collapse(
                        adjust_collapsed_head_children_edge_lengths=not args.discardLengths
                    )

        # Re-root.
        if args.rootNode:
            node = tree.find_node_with_taxon_label(args.rootNode)
            assert node is not None
            tree.reroot_at_edge(
                node.edge,
                update_bipartitions=True,
                length1=node.edge.length / args.scale,
                length2=node.edge.length / args.scale,
            )

        if args.ladderize:
            # The reorder function is from dendropy v5.0.2 (released Sept 2024).
            tree.reorder()
            tree.ladderize(args.ladderize == "ascending")

        if args.dryRun:
            print(f"Would write curated tree file {str(curated)!r}.")
        else:
            tree.write(path=curated, schema=args.format)
            print(f"Wrote curated tree file {str(curated)!r}.")


if __name__ == "__main__":
    main()



from unittest import TestCase
from os.path import dirname, join

import dark
from dark.genomes import GenomeProteinInfo
from dark.civ.proteins import SqliteIndex

TOP = dirname(dirname(dark.__file__))

DB = SqliteIndex(join(TOP, "test", "data", "hbv", "hbv-proteins.db"))

BAM1 = join(TOP, "test", "data", "hbv", "query1.bam")
BAM2 = join(TOP, "test", "data", "hbv", "query2.bam")
BAM3 = join(TOP, "test", "data", "hbv", "query3.bam")


class TestGenomeProteinInfo(TestCase):
    """
    Test the GenomeProteinInfo class.
    """

    def testLoadReference(self):
        """
        Test that everything is as expected after loading the genome file.
        """
        gpi = GenomeProteinInfo("KJ586809.1", DB, True)
        self.assertEqual(gpi.genome["proteinCount"], len(gpi.proteins))
        nonProteinOffsets = set(range(gpi.genome["length"])) - set(gpi.offsets)
        self.assertEqual(set(), nonProteinOffsets)
        self.assertEqual("KJ586809.1", gpi.genome["accession"])
        self.assertEqual(
            "Hepatitis B virus strain P18, complete genome", gpi.genome["name"]
        )

    def testLoadBAM1(self):
        """
        Test that everything is as expected after loading the BAM1 file.
        """
        gpi = GenomeProteinInfo("KJ586809.1", DB, True)
        gpi.addSAM(BAM1)

        # Genome covered offsets.
        self.assertEqual(200, len(gpi.coveredOffsetCount))
        self.assertEqual(list(range(200)), list(gpi.coveredOffsetCount))
        self.assertEqual([1] * 200, list(gpi.coveredOffsetCount.values()))

        # SAM files.
        self.assertEqual([BAM1], gpi.samFiles)
        self.assertEqual({"query1"}, gpi.readIdsMatchingGenome)

        # Protein accession numbers.
        expected = set(["AJF208%02d.1" % i for i in range(4, 11)])
        self.assertEqual(expected, set(gpi.proteins))

        # Offset 200 is in 4 proteins but is not matched by the query.
        self.assertEqual(set(), gpi.offsets[200]["readIds"])
        expected = set(["AJF208%02d.1" % i for i in range(4, 8)])
        self.assertEqual(expected, gpi.offsets[200]["proteinAccessions"])

        # Offset 0 is in 3 proteins and is matched by the query.
        self.assertEqual({"query1"}, gpi.offsets[0]["readIds"])
        expected = set(["AJF208%02d.1" % i for i in range(4, 7)])
        self.assertEqual(expected, gpi.offsets[0]["proteinAccessions"])

        # Read ids for all proteins.
        self.assertEqual({"query1"}, gpi.readIdsForAllProteins())

        # AJF20804.1 coverage (its ranges are 2306-3221 and 0-1623)
        info = gpi.proteinCoverageInfo("AJF20804.1")
        self.assertEqual(200, info["coveredOffsets"])
        self.assertEqual(200, info["totalBases"])
        self.assertEqual((3221 - 2306) + (1623 - 0), info["ntLength"])
        self.assertEqual({"query1"}, info["readIds"])

    def testLoadBAM12(self):
        """
        Test that everything is as expected after loading the BAM1 and BAM2
        files.
        """
        gpi = GenomeProteinInfo("KJ586809.1", DB, True)
        gpi.addSAM(BAM1)
        gpi.addSAM(BAM2)

        # Genome covered offsets.
        self.assertEqual(300, len(gpi.coveredOffsetCount))
        self.assertEqual(
            list(range(200)) + list(range(1400, 1500)), list(gpi.coveredOffsetCount)
        )
        self.assertEqual([1] * 300, list(gpi.coveredOffsetCount.values()))

        # SAM files.
        self.assertEqual([BAM1, BAM2], gpi.samFiles)
        self.assertEqual({"query1", "query2"}, gpi.readIdsMatchingGenome)

        # Protein accession numbers.
        expected = set(["AJF208%02d.1" % i for i in range(4, 11)])
        self.assertEqual(expected, set(gpi.proteins))

        # Offset 200 is in 4 proteins but is not matched by any query.
        self.assertEqual(set(), gpi.offsets[200]["readIds"])
        expected = set(["AJF208%02d.1" % i for i in range(4, 8)])
        self.assertEqual(expected, gpi.offsets[200]["proteinAccessions"])

        # Offset 0 is in 3 proteins and is matched by query1.
        self.assertEqual({"query1"}, gpi.offsets[0]["readIds"])
        expected = set(["AJF208%02d.1" % i for i in range(4, 7)])
        self.assertEqual(expected, gpi.offsets[0]["proteinAccessions"])

        # Offset 1400 is in 2 proteins and is matched by query2.
        self.assertEqual({"query2"}, gpi.offsets[1400]["readIds"])
        self.assertEqual(
            {"AJF20804.1", "AJF20808.1"}, gpi.offsets[1400]["proteinAccessions"]
        )

        # Read ids for all proteins.
        self.assertEqual({"query1", "query2"}, gpi.readIdsForAllProteins())

        # AJF20804.1 coverage (its ranges are 2306-3221 and 0-1623)
        info = gpi.proteinCoverageInfo("AJF20804.1")
        self.assertEqual(300, info["coveredOffsets"])
        self.assertEqual(300, info["totalBases"])
        self.assertEqual((3221 - 2306) + (1623 - 0), info["ntLength"])
        self.assertEqual({"query1", "query2"}, info["readIds"])

    def testLoadBAM123(self):
        """
        Test that everything is as expected after loading the BAM1, BAM2,
        and BAM3 files.
        """
        gpi = GenomeProteinInfo("KJ586809.1", DB, True)
        gpi.addSAM(BAM1)
        gpi.addSAM(BAM2)
        gpi.addSAM(BAM3)

        # Genome covered offsets.
        # There are 50 offsets that are covered twice.
        self.assertEqual(750 - 50, len(gpi.coveredOffsetCount))
        self.assertEqual(
            set(range(200)) | set(range(1000, 1500)), set(gpi.coveredOffsetCount)
        )
        self.assertEqual(
            set([1] * 700 + [2] * 50), set(gpi.coveredOffsetCount.values())
        )

        # SAM files.
        self.assertEqual([BAM1, BAM2, BAM3], gpi.samFiles)
        self.assertEqual({"query1", "query2", "query3"}, gpi.readIdsMatchingGenome)

        # Protein accession numbers.
        expected = set(["AJF208%02d.1" % i for i in range(4, 11)])
        self.assertEqual(expected, set(gpi.proteins))

        # Offset 200 is in 4 proteins but is not matched by any query.
        self.assertEqual(set(), gpi.offsets[200]["readIds"])
        expected = set(["AJF208%02d.1" % i for i in range(4, 8)])
        self.assertEqual(expected, gpi.offsets[200]["proteinAccessions"])

        # Offset 0 is in 3 proteins and is matched by query1.
        self.assertEqual({"query1"}, gpi.offsets[0]["readIds"])
        expected = set(["AJF208%02d.1" % i for i in range(4, 7)])
        self.assertEqual(expected, gpi.offsets[0]["proteinAccessions"])

        # Offset 1400 is in 2 proteins and is matched by query2 and query3.
        self.assertEqual({"query2", "query3"}, gpi.offsets[1400]["readIds"])
        self.assertEqual(
            {"AJF20804.1", "AJF20808.1"}, gpi.offsets[1400]["proteinAccessions"]
        )

        # Read ids for all proteins.
        self.assertEqual({"query1", "query2", "query3"}, gpi.readIdsForAllProteins())

        # AJF20804.1 coverage (its ranges are 2306-3221 and 0-1623)
        info = gpi.proteinCoverageInfo("AJF20804.1")
        self.assertEqual(700, info["coveredOffsets"])
        self.assertEqual(750, info["totalBases"])
        self.assertEqual((3221 - 2306) + (1623 - 0), info["ntLength"])
        self.assertEqual({"query1", "query2", "query3"}, info["readIds"])

    def testTooFewReadOffsetsBAM1(self):
        """
        Test that a read is not returned as overlapping a protein unless it
        meets the minimum number of required overlapping offsets.
        """
        gpi = GenomeProteinInfo("KJ586809.1", DB, True)
        gpi.addSAM(BAM1)

        # Look at protein AJF20804.1 coverage (its ranges are 2306-3221 and
        # 0-1623). There should be no matching reads because the query
        # (query1) is only 200 nt long and so cannot match with at least
        # 500 nucleotides. The number of covered offsets and total bases
        # should both also be zero for the same reason.
        info = gpi.proteinCoverageInfo("AJF20804.1", 500)
        self.assertEqual(set(), info["readIds"])
        self.assertEqual(0, info["totalBases"])
        self.assertEqual(0, info["coveredOffsets"])

    def testSufficientReadOffsetsBAM1(self):
        """
        Test that a read is returned as overlapping a protein when it meets
        the minimum number of required overlapping offsets.
        """
        gpi = GenomeProteinInfo("KJ586809.1", DB, True)
        gpi.addSAM(BAM1)

        # Look at protein AJF20804.1 coverage (its ranges are 2306-3221 and
        # 0-1623). The query (query1) must be returned as it has 200
        # matching nucleotides.
        info = gpi.proteinCoverageInfo("AJF20804.1", 199)
        self.assertEqual({"query1"}, info["readIds"])



#!/usr/bin/env python

"""
This script needs work!  Or will we never use it again?

There should be a --find option, or something.
"""

from dark.blast import BlastRecords
from dark.graphics import evalueGraph

if __name__ == "__main__":
    import sys

    if len(sys.argv) < 2:
        print("Usage: %s BLAST-hitfile..." % sys.argv[0], file=sys.stderr)
    else:
        blastRecords = BlastRecords(sys.argv[1:])
        evalueGraph(
            blastRecords.records(),
            10,
            10,
            find=lambda title: title.find("HKU4") > -1
            and title.find("complete genome") > -1,
            titles=False,
        )



from os.path import exists
from threading import RLock
from cachetools import LFUCache, cached


class _FilePointerCache(LFUCache):
    """
    A least-frequently-used file pointer cache class with an eviction method
    that closes open files.
    """

    def popitem(self):
        """
        Evict a cache item.

        @return: The C{str} name of the evicted file.
        """
        _, (filename, fp) = super().popitem()
        fp.close()
        return filename

    def close(self):
        """
        Close all open files in the cache.
        """
        while self.currsize:
            self.popitem()


class FilePointerCache:
    """
    A file pointer cache class for simultaneously opening multiple files,
    without exceeeding the per-process operating system limit on open file
    handles.

    @param maxsize: The C{int} maximum size of the cache.
    @param openArgs: If not C{None}, a C{dict} of keyword arguments to pass to
        open when opening a new file. If C{None}, the file will be opened
        with the default mode ('rt').
    @param reopenArgs: If not C{None}, a C{dict} of keyword arguments to pass
        to open when opening an already existing file. If C{None}, the file
        will be opened with the default mode ('rt').
    """

    def __init__(self, maxsize=32, openArgs=None, reopenArgs=None):
        self._openArgs = openArgs or {}
        self._reopenArgs = reopenArgs or {}
        self._filePointerCache = _FilePointerCache(maxsize=maxsize)
        self._lock = RLock()

        @cached(self._filePointerCache, lock=self._lock)
        def _open(filename):
            if exists(filename):
                fp = open(filename, **self._reopenArgs)
            else:
                fp = open(filename, **self._openArgs)
            return filename, fp

        self._open = _open

    def open(self, filename):
        return self._open(filename)[1]

    def close(self):
        with self._lock:
            self._filePointerCache.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()



from typing import Union, Optional
from io import TextIOWrapper
from collections import defaultdict
from Bio import SeqIO

from dark.reads import Read
from dark.utils import median


def summarizeReads(
    file_handle: Union[str, TextIOWrapper], file_type: str
) -> dict[str, Union[int, float, defaultdict[str, int]]]:
    """
    open a fasta or fastq file, prints number of of reads,
    average length of read, total number of bases, longest,
    shortest and median read, total number and average of
    individual base (A, T, G, C, N).
    """
    base_counts: defaultdict[str, int] = defaultdict(int)
    read_number = 0
    total_length = 0
    length_list: list[int] = []

    records = SeqIO.parse(file_handle, file_type)

    for record in records:
        total_length += len(record)
        read_number += 1
        length_list.append(len(record))
        for base in record:
            base_counts[base] += 1

    result: dict[str, Union[int, float, defaultdict[str, int]]] = {
        "read_number": read_number,
        "total_length": total_length,
        "average_length": total_length / read_number if read_number > 0 else 0,
        "max_length": max(length_list) if length_list else 0,
        "min_length": min(length_list) if length_list else 0,
        "median_length": median(length_list) if length_list else 0,
        "base_counts": base_counts,
    }

    return result


def sequenceCategoryLengths(
    read: Read,
    categories: dict[str, str],
    defaultCategory: Optional[str] = None,
    suppressedCategory: str = "...",
    minLength: int = 1,
):
    """
    Summarize the nucleotides or AAs found in a read by assigning each to a
    category and reporting the lengths of the contiguous category classes
    found along the sequence.

    @param read: A C{Read} instance or one of its subclasses.
    @param categories: A C{dict} mapping nucleotides or AAs to category.
    @param defaultCategory: The category to use if a sequence base is not
        in C{categories}.
    @param suppressedCategory: The category to use to indicate suppressed
        sequence regions (i.e., made up of stretches of bases that are less
        than C{minLength} in length).
    @param minLength: stretches of the read that are less than this C{int}
        length will be summed and reported as being in the
        C{suppressedCategory} category.
    @raise ValueError: If minLength is less than one.
    @return: A C{list} of 2-C{tuples}. Each tuple contains a (category, count)
        where category is either a C{str} or C{None} and count is an C{int}.
    """
    result: list[tuple[Optional[str], int]] = []
    append = result.append
    get = categories.get
    first = True
    currentCategory = None
    currentCount = 0
    suppressing = False
    suppressedCount = 0

    if minLength < 1:
        raise ValueError("minLength must be at least 1")

    for base in read.sequence:
        thisCategory = get(base, defaultCategory)
        if first:
            first = False
            currentCategory = thisCategory
            currentCount += 1
        else:
            if thisCategory == currentCategory:
                # This base is still in the same category as the last base.
                # Keep counting.
                currentCount += 1
            else:
                # This is a new category.
                if currentCount < minLength:
                    # The category region that was just seen will not be
                    # emitted.
                    if suppressing:
                        # Already suppressing. Suppress the just-seen region too.
                        suppressedCount += currentCount
                    else:
                        # Start suppressing.
                        suppressedCount = currentCount
                        suppressing = True
                else:
                    if suppressing:
                        append((suppressedCategory, suppressedCount))
                        suppressedCount = 0
                        suppressing = False
                    append((currentCategory, currentCount))
                currentCategory = thisCategory
                currentCount = 1

    if suppressing:
        append((suppressedCategory, suppressedCount + currentCount))
    elif currentCount >= minLength:
        append((currentCategory, currentCount))
    elif currentCount:
        append((suppressedCategory, currentCount))

    return result



#!/usr/bin/env python

from Bio import SeqIO
from dark.analyze_reads import getPrefixAndSuffix
from dark.analyze_reads import trimReads
import sys


if len(sys.argv) != 2:
    print("getPrefixAndSuffix() takes exactly 1 argument", file=sys.stderr)
    sys.exit(1)

else:
    filename = sys.argv[1]
    prefix, suffix = getPrefixAndSuffix(filename)

    print("Prefix length %d, suffix length %d" % (prefix, suffix), file=sys.stderr)

    SeqIO.write(trimReads(prefix, suffix, filename), sys.stdout, "fasta")



from typing import IO, Sequence, AnyStr

from Bio import SeqIO


def _longestPrefixOfTwoSeqs(a: str, b: str):
    length = min(len(a), len(b))
    result = 0
    while result < length:
        if a[result] == b[result]:
            result += 1
        else:
            break
    return result


def getPrefixAndSuffix(file_handle: IO[AnyStr]):
    read_list = list(SeqIO.parse(file_handle, "fasta"))
    reversed_read_list = [read[::-1] for read in read_list]

    def longestCommonPrefix(read_list: Sequence):
        sequences = read_list
        nSequences = len(sequences)
        if nSequences == 1:
            return len(sequences[0])
        elif nSequences == 0:
            return 0
        else:
            prefix = sequences[0]
            result = len(prefix)
            index = 1
            while index < nSequences:
                thisLen = _longestPrefixOfTwoSeqs(prefix, sequences[index])
                if thisLen == 0:
                    return 0
                elif thisLen < result:
                    prefix = prefix[:thisLen]
                    result = thisLen
                index += 1
            return result

    prefix = longestCommonPrefix(read_list)
    suffix = longestCommonPrefix(reversed_read_list)
    return prefix, suffix


def trimReads(prefix: int, suffix: int, file_handle: IO[AnyStr]):
    for record in SeqIO.parse(file_handle, "fasta"):
        if suffix == 0:
            yield record[prefix:]
        else:
            yield record[prefix:-suffix]



#!/usr/bin/env python

import argparse
import bz2file
import sys

from dark.diamond.conversion import DiamondTabularFormatReader


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert DIAMOND tabular output to JSON.",
        epilog=(
            "Give a DIAMOND tabular file and convert it to JSON, "
            "optionally compressing the output. You *must* invoke "
            "DIAMOND with the following output specification: "
            "--outfmt 6 qtitle stitle bitscore evalue qframe qseq "
            "qstart qend sseq sstart send slen btop"
            "Note that only each line of the output is JSON (the full "
            "output is not valid JSON by itself)."
        ),
    )

    parser.add_argument(
        "--json",
        metavar="JSON-output-file",
        help=(
            "The JSON filename to write the converted DIAMOND file to. If "
            "omitted, standard output will be used."
        ),
    )

    parser.add_argument(
        "--diamond",
        metavar="DIAMOND-file",
        default=sys.stdin,
        help=(
            "The DIAMOND tabular output file to convert. If omitted, "
            "standard input will be read."
        ),
    )

    parser.add_argument(
        "--bzip2",
        default=False,
        action="store_true",
        help="Compress output using bzip2.",
    )

    args = parser.parse_args()

    if args.bzip2:
        fp = bz2file.BZ2File(args.json or sys.stdout, "w")
    else:
        fp = open(args.json, "w") if args.json else sys.stdout

    reader = DiamondTabularFormatReader(args.diamond)
    reader.saveAsJSON(fp, writeBytes=args.bzip2)
    fp.close()



#!/usr/bin/env python

import sys
import argparse
import re

from dark.taxonomy import (
    addTaxonomyDatabaseCommandLineOptions,
    parseTaxonomyDatabaseCommandLineOptions,
)

VERSION_REGEX = re.compile(r"\.\d+$")


def getHosts(id_, db) -> set[str]:
    """
    Look up hosts for an accession or taxonomy id.

    @param id_: The C{str} accession number or taxonomy id to look up.
    @param db: The sqlite3 taxonomy database to consult.
    @return: A C{set} of C{str} host names.
    """
    try:
        idInt = int(id_)
    except ValueError:
        hosts = db.hosts(id_)
    else:
        hosts = db.hosts(idInt)

    if hosts is None and VERSION_REGEX.search(id_) is None:
        # Try adding a version number.
        hosts = db.hosts(id_ + ".1")

    return hosts


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Print hosts for accession numbers or taxonomy ids",
    )

    parser.add_argument(
        "ids",
        nargs="*",
        help=(
            "The ids (accession numbers, names, or taxonomy ids) to print "
            "host information for. If not given, ids are read from "
            "standard input, one per line."
        ),
    )

    parser.add_argument(
        "--database",
        required=True,
        help=(
            "The file holding the sqlite3 taxonomy database. See "
            "https://github.com/acorg/ncbi-taxonomy-database for how to "
            "build one."
        ),
    )

    parser.add_argument(
        "--printId",
        default=False,
        action="store_true",
        help="If specified, also print the id.",
    )

    addTaxonomyDatabaseCommandLineOptions(parser)

    args = parser.parse_args()

    db = parseTaxonomyDatabaseCommandLineOptions(args, parser)

    if args.ids:
        ids = args.ids
    else:
        ids = (line[:-1] for line in sys.stdin)

    for id_ in ids:
        if args.printId:
            print(id_ + ":")
        hosts = getHosts(id_, db)
        if hosts:
            print(", ".join(sorted(hosts)))
        else:
            print(
                "No host information for %r found in the taxonomy database." % id_,
                file=sys.stderr,
            )



import builtins
from io import StringIO

from dark.reads import AARead, DNARead, RNARead
from dark.fastq import FastqReads

from unittest import TestCase, skip
from unittest.mock import patch, mock_open


class TestFastqReads(TestCase):
    """
    Tests for the L{dark.fastq.FastqReads} class.
    """

    def testEmpty(self):
        """
        An empty FASTQ file results in an empty iterator.
        """
        with patch.object(builtins, "open", mock_open()):
            reads = FastqReads("filename.fastq")
            self.assertEqual([], list(reads))

    def testOneRead(self):
        """
        A FASTQ file with one read must be read properly.
        """
        data = "\n".join(["@id1", "ACGT", "+", "!!!!"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(FastqReads("filename.fastq"))
            self.assertEqual([DNARead("id1", "ACGT", "!!!!")], reads)

    def testTwoReads(self):
        """
        A FASTQ file with two reads must be read properly and its
        sequences must be returned in the correct order.
        """
        data = "\n".join(["@id1", "ACGT", "+", "!!!!", "@id2", "TGCA", "+", "????"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(FastqReads("filename.fastq"))
            self.assertEqual(2, len(reads))
            self.assertEqual(
                [DNARead("id1", "ACGT", "!!!!"), DNARead("id2", "TGCA", "????")], reads
            )

    def testTypeDefaultsToDNA(self):
        """
        A FASTQ file whose type is not specified must result in reads that
        are instances of DNARead.
        """
        data = "\n".join(["@id1", "ACGT", "+", "!!!!"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(FastqReads("filename.fastq"))
            self.assertTrue(isinstance(reads[0], DNARead))

    def testTypeAA(self):
        """
        A FASTQ file whose read class is AARead must result in reads that
        are instances of AARead.
        """
        data = "\n".join(["@id1", "ACGT", "+", "!!!!"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(FastqReads("filename.fastq", AARead))
            self.assertTrue(isinstance(reads[0], AARead))

    def testTypeDNA(self):
        """
        A FASTQ file whose read class is DNARead must result in reads that
        are instances of DNARead.
        """
        data = "\n".join(["@id1", "ACGT", "+", "!!!!"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(FastqReads("filename.fastq", DNARead))
            self.assertTrue(isinstance(reads[0], DNARead))

    def testTypeRNA(self):
        """
        A FASTQ file whose read class is RNARead must result in reads that
        are instances of RNARead.
        """
        data = "\n".join(["@id1", "ACGT", "+", "!!!!"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(FastqReads("filename.fastq", RNARead))
            self.assertTrue(isinstance(reads[0], RNARead))

    @skip("Some tests are broken and skipped under latest BioPython")
    def testTwoFiles(self):
        """
        It must be possible to read from two FASTQ files.
        """

        class SideEffect:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename):
                if self.count == 0:
                    self.test.assertEqual("file1.fastq", filename)
                    self.count += 1
                    return StringIO("@id1\nACTG\n+\n!!!!\n")
                elif self.count == 1:
                    self.test.assertEqual("file2.fastq", filename)
                    self.count += 1
                    return StringIO("@id2\nCAGT\n+\n!!!!\n")
                else:
                    self.test.fail("We are only supposed to be called twice!")

        sideEffect = SideEffect(self)
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect.sideEffect
            reads = FastqReads(["file1.fastq", "file2.fastq"])
            self.assertEqual(
                [
                    DNARead("id1", "ACTG", "!!!!"),
                    DNARead("id2", "CAGT", "!!!!"),
                ],
                list(reads),
            )



#!/usr/bin/env python

import sys

from dark.reads import DNARead


if len(sys.argv) < 2:
    print("Usage: reverse-complement.py SEQ1 [SEQ2, SEQ3, ...]", file=sys.stderr)
    sys.exit(1)
else:
    for seq in sys.argv[1:]:
        print(DNARead("id", seq).reverseComplement().sequence)



#!/usr/bin/env python

import sys
import argparse
import re

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions, Reads

parser = argparse.ArgumentParser(
    description=(
        "Write sorted FASTA/Q to stdout. Sorting is by sequence id, "
        "then by sequence, then by quality (if FASTQ)"
    )
)

# TODO: Support multiple regexes and apply them one by one to allow matching on
# different regions in an arbitrary order.
parser.add_argument(
    "--regex",
    help=(
        "A regular expression to specify a region (or regions) to extract "
        "from sequence ids to sort on. Regions will be converted to "
        "integers if possible."
    ),
)


parser.add_argument(
    "--reverse", "-r", action="store_true", help="Sort by decreasing value."
)

addFASTACommandLineOptions(parser)
args = parser.parse_args()
reads = parseFASTACommandLineOptions(args)

if args.regex:
    regex = re.compile(args.regex)

    if regex.groups == 0:
        print(
            "You have passed a regular expression that has no capturing "
            "group(s) specified using (...).",
            file=sys.stderr,
        )
        sys.exit(1)

    def key(read):
        result = []
        match = regex.search(read.id)
        if match:
            for text in match.groups():
                try:
                    value = int(text)
                except ValueError:
                    value = text
                result.append(value)
        return result

else:

    def key(read):
        return read.id


Reads(sorted(reads, key=key, reverse=args.reverse)).save(
    sys.stdout, "fastq" if args.fastq else "fasta"
)



from Bio import Entrez, SeqIO
from urllib import error as URLError

Entrez.email = "tcj25@cam.ac.uk"  # type: ignore


def getSequence(title, db="nucleotide"):
    """
    Get information about a sequence from Genbank.

    @param title: A C{str} sequence title from a BLAST hit. Of the form
        'gi|63148399|gb|DQ011818.1| Description...'.
    @param db: The C{str} name of the Entrez database to consult.

    NOTE: this uses the network!  Also, there is a 3 requests/second limit
    imposed by NCBI on these requests so be careful or your IP will be banned.
    """
    titleId = title.split(" ", 1)[0]
    try:
        gi = titleId.split("|")[1]
    except IndexError:
        # Assume we have a gi number directly, and make sure it's a string.
        gi = str(titleId)

    try:
        client = Entrez.efetch(db=db, rettype="gb", retmode="text", id=gi)
    except URLError:
        return None
    else:
        record = SeqIO.read(client, "gb")
        client.close()
        return record



# Import and make available things that are convenient to have around in
# iPythonNotebook following 'from dark.ipynb import *'.

from .fasta import FastaReads
from .blast.alignments import BlastReadsAlignments
from .titles import titleCounts, TitlesAlignments
from .html import (
    summarizeTitlesByLength,
    summarizeTitlesByMaxScore,
    summarizeTitlesByMedianScore,
    summarizeTitlesByCount,
    summarizeTitlesByTitle,
)
from .graphics import alignmentGraph, alignmentPanel, scoreGraph

# Keep pyflakes quiet by pretending to make use of all our imports.
_ = (
    FastaReads,
    BlastReadsAlignments,
    titleCounts,
    TitlesAlignments,
    summarizeTitlesByLength,
    summarizeTitlesByMaxScore,
    summarizeTitlesByMedianScore,
    summarizeTitlesByCount,
    summarizeTitlesByTitle,
    alignmentGraph,
    alignmentPanel,
    scoreGraph,
)

__all__ = [
    "FastaReads",
    "BlastReadsAlignments",
    "titleCounts",
    "TitlesAlignments",
    "summarizeTitlesByLength",
    "summarizeTitlesByMaxScore",
    "summarizeTitlesByMedianScore",
    "summarizeTitlesByCount",
    "summarizeTitlesByTitle",
    "alignmentGraph",
    "alignmentPanel",
    "scoreGraph",
]



#!/usr/bin/env python

"""
Read DNA FASTA from stdin and print FASTA to stdout, with the sequences
randomized. The read ids and lengths are preserved.

Note: This produces DNA sequences. If you have AA reads and you need this
      functionality, we can add it.
"""

import sys
from numpy.random import choice

from dark.fasta import FastaReads
from dark.reads import DNARead


if __name__ == "__main__":
    for read in FastaReads(sys.stdin):
        seq = "".join(choice(["A", "C", "G", "T"], len(read), replace=True))
        print(DNARead(read.id, seq).toString("fasta"))



from hashlib import md5
import sqlite3
import os
from typing import Generator, Iterator, Iterable, TextIO, Union, Type, Optional, List

from Bio import SeqIO, bgzf  # type: ignore

from dark.reads import Reads, DNARead, Read
from dark.sqlite3 import sqliteConnect
from dark.utils import asHandle


def fastaToList(fastaFilename) -> list:
    return list(SeqIO.parse(fastaFilename, "fasta"))


def dedupFasta(reads: Reads) -> Generator[Read, None, None]:
    """
    Remove sequence duplicates (based on sequence) from FASTA.

    @param reads: a C{dark.reads.Reads} instance.
    @return: a generator of C{dark.reads.Read} instances with no duplicates.
    """
    seen: set[bytes] = set()
    add = seen.add

    for read in iter(reads):
        hash_ = md5(read.sequence.encode("UTF-8")).digest()
        if hash_ not in seen:
            add(hash_)
            yield read


def dePrefixAndSuffixFasta(
    sequences: Iterable[SeqIO.SeqRecord],
) -> Iterator[SeqIO.SeqRecord]:
    """
    sequences: an iterator producing Bio.Seq sequences.

    @return: a generator of sequences with no duplicates and no fully contained
        subsequences.
    """
    sequences = sorted(sequences, key=lambda s: len(s.seq), reverse=True)
    seen = set()
    for s in sequences:
        thisSeq = str(s.seq)
        thisHash = md5(thisSeq.encode("UTF-8")).digest()
        if thisHash not in seen:
            # Add prefixes.
            newHash = md5()
            for nucl in thisSeq:
                newHash.update(nucl.encode("UTF-8"))
                seen.add(newHash.digest())
            # Add suffixes.
            for start in range(len(thisSeq) - 1):
                seen.add(md5(thisSeq[start + 1 :].encode("UTF-8")).digest())
            yield s


def fastaSubtract(fastaFiles: List[TextIO]) -> Iterator[SeqIO.SeqRecord]:
    """
    Given a list of open file descriptors, each with FASTA content,
    remove the reads found in the 2nd, 3rd, etc files from the first file
    in the list.

    @param fastaFiles: a C{list} of FASTA filenames.
    @raises IndexError: if passed an empty list.
    @return: An iterator producing C{Bio.Seq} instances suitable for
        writing to a file using C{Bio.SeqIO.write}.

    """
    reads = {}
    fastaFiles = list(fastaFiles)
    firstFile = fastaFiles.pop(0)
    for seq in SeqIO.parse(firstFile, "fasta"):
        reads[seq.id] = seq

    for fastaFile in fastaFiles:
        for seq in SeqIO.parse(fastaFile, "fasta"):
            # Make sure that reads with the same id have the same sequence.
            if seq.id in reads:
                assert str(seq.seq) == str(reads[seq.id].seq)
            reads.pop(seq.id, None)

    return iter(reads.values())


class FastaReads(Reads):
    """
    Subclass of L{dark.reads.Reads} providing access to FASTA reads.

    @param _files: Either a single C{str} file name or file handle, or a
        C{list} of C{str} file names and/or file handles. Each file or file
        handle must contain sequences in FASTA format.
    @param readClass: The class of read that should be yielded by iter.
    @param upperCase: If C{True}, read sequences will be converted to upper
        case.
    """

    def __init__(
        self,
        _files: Union[str, TextIO, Union[list, tuple]],
        readClass: Type[Read] = DNARead,
        upperCase: bool = False,
    ):
        self._files = _files if isinstance(_files, (list, tuple)) else [_files]
        self._readClass = readClass
        # TODO: It would be better if upperCase were an argument that could
        # be passed to Reads.__init__ and that could do the uppercasing in
        # its add method (as opposed to using it below in our iter method).
        # In that case, in the iter of this class we'd call self.add on
        # each of the sequences coming from self._file. Or, if we'd already
        # read the file we'd return Reads.iter(self) to re-iterate over the
        # sequences already added from the file.
        self._upperCase = upperCase
        super().__init__()

    def iter(self) -> Generator[Read, None, None]:
        """
        Iterate over the sequences in the files in self.files_, yielding each
        as an instance of the desired read class.
        """
        for _file in self._files:
            with asHandle(_file) as fp:
                # Duplicate some code here so as not to test
                # self._upperCase in the loop.
                if self._upperCase:
                    for seq in SeqIO.parse(fp, "fasta"):
                        read = self._readClass(seq.description, str(seq.seq.upper()))
                        yield read
                else:
                    for seq in SeqIO.parse(fp, "fasta"):
                        read = self._readClass(seq.description, str(seq.seq))
                        yield read


def combineReads(
    filename: str,
    sequences: list[str],
    readClass: Type[Read] = DNARead,
    upperCase: bool = False,
    idPrefix: str = "command-line-read-",
) -> Union[FastaReads, Reads]:
    """
    Combine FASTA reads from a file and/or sequence strings.

    @param filename: A C{str} file name containing FASTA reads.
    @param sequences: A C{list} of C{str} sequences. If a sequence
        contains spaces, the last field (after splitting on spaces) will be
        used as the sequence and the first fields will be used as the sequence
        id.
    @param readClass: The class of the individual reads.
    @param upperCase: If C{True}, reads will be converted to upper case.
    @param idPrefix: The C{str} prefix that will be used for the id of the
        sequences in C{sequences} that do not have an id specified. A trailing
        sequence number will be appended to this prefix. Note that
        'command-line-read-', the default id prefix, could collide with ids in
        the FASTA file, if given. So output might be ambiguous. That's why we
        allow the caller to specify a custom prefix.
    @return: A C{FastaReads} instance.
    """
    # Read sequences from a FASTA file, if given.
    reads: Union[FastaReads, Reads]

    if filename:
        reads = FastaReads(filename, readClass=readClass, upperCase=upperCase)
    else:
        reads = Reads()

    # Add any individually specified subject sequences.
    if sequences:
        for count, sequence in enumerate(sequences, start=1):
            # Try splitting the sequence on its last space and using the
            # first part of the split as the read id. If there's no space,
            # assign a generic id.
            parts = sequence.rsplit(" ", 1)
            if len(parts) == 2:
                readId, sequence = parts
            else:
                readId = "%s%d" % (idPrefix, count)
            if upperCase:
                sequence = sequence.upper()
            read = readClass(readId, sequence)
            reads.add(read)

    return reads


class SqliteIndex:
    """
    Create an Sqlite3 database holding FASTA sequence ids, file names, and
    offsets for fast random dictionary-like access.

    @param dbFilename: A C{str} file name containing an sqlite3 database. If
        the file does not exist it will be created. The special string
        ":memory:" can be used to create an in-memory database.
    @param readClass: The class of read that should be returned by __getitem__.
    @param fastaDirectory: A C{str} directory where the indexed FASTA files
        can be found. If provided, this directory is only used by __getitem__,
        which will combine it with the basename of the files given to
        C{addFile} to locate the FASTA.
    """

    def __init__(
        self,
        dbFilename: str,
        readClass: Type[Read] = DNARead,
        fastaDirectory: Optional[str] = None,
    ):
        self._readClass = readClass
        self._fastaDirectory = fastaDirectory
        creating = dbFilename == ":memory:" or not os.path.exists(dbFilename)
        self._connection: sqlite3.Connection = sqliteConnect(dbFilename)
        if creating:
            # Create a new database.
            cur = self._connection.cursor()
            cur.executescript(
                """
                CREATE TABLE files (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    name VARCHAR UNIQUE
                );

                CREATE TABLE sequences (
                    id VARCHAR UNIQUE PRIMARY KEY,
                    fileNumber INTEGER,
                    offset INTEGER
                );
            """
            )
            self._connection.commit()

    def _getFilename(self, fileNumber: int) -> Optional[str]:
        """
        Given a file number, get its name (if any).

        @param fileNumber: An C{int} file number.
        @return: A C{str} file name or C{None} if a file with that number
            has not been added.
        """
        cur = self._connection.cursor()
        cur.execute("SELECT name FROM files WHERE id = ?", (fileNumber,))
        row = cur.fetchone()
        if row is None:
            return None
        else:
            return row[0]

    def _getFileNumber(self, filename: str) -> Optional[int]:
        """
        Given a file name, get its file number (if any).

        @param filename: A C{str} file name.
        @return: An C{int} file number or C{None} if no file with that name
            has been added.
        """
        cur = self._connection.cursor()
        cur.execute("SELECT id FROM files WHERE name = ?", (filename,))
        row = cur.fetchone()
        if row is None:
            return None
        else:
            return row[0]

    def _addFilename(self, filename: str) -> Optional[int]:
        """
        Add a new file name.

        @param filename: A C{str} file name.
        @raise ValueError: If a file with this name has already been added.
        @return: The C{int} id of the newly added file.
        """
        cur = self._connection.cursor()
        try:
            cur.execute("INSERT INTO files(name) VALUES (?)", (filename,))
        except sqlite3.IntegrityError as e:
            if str(e).find("UNIQUE constraint failed") > -1:
                raise ValueError("Duplicate file name: %r" % filename)
            else:
                raise
        else:
            fileNumber = cur.lastrowid
            self._connection.commit()
            return fileNumber

    def addFile(self, filename: str) -> int:
        """
        Add a new FASTA file of sequences.

        @param filename: A C{str} file name, with the file in FASTA format.
            This file must (obviously) exist at indexing time. When __getitem__
            is used to access sequences, it is possible to provide a
            C{fastaDirectory} argument to our C{__init__} to indicate the
            directory containing the original FASTA files, in which case the
            basename of the file here provided in C{filename} is used to find
            the file in the given directory. This allows the construction of a
            sqlite database from the shell in one directory and its use
            programmatically from another directory.
        @raise ValueError: If a file with this name has already been added or
            if the file contains a sequence whose id has already been seen.
        @return: The C{int} number of sequences added from the file.
        """
        endswith = filename.lower().endswith
        if endswith(".bgz") or endswith(".gz"):
            useBgzf = True
        elif endswith(".bz2"):
            raise ValueError(
                "Compressed FASTA is only supported in BGZF format. Use "
                "bgzip to compresss your FASTA."
            )
        else:
            useBgzf = False

        fileNumber = self._addFilename(filename)
        connection = self._connection
        count = 0
        try:
            with connection:
                if useBgzf:
                    try:
                        fp = bgzf.open(filename, "rb")
                    except ValueError as e:
                        if str(e).find("BGZF") > -1:
                            raise ValueError(
                                "Compressed FASTA is only supported in BGZF "
                                "format. Use the samtools bgzip utility "
                                "(instead of gzip) to compresss your FASTA."
                            )
                        else:
                            raise
                    else:
                        try:
                            for line in fp:
                                if line[0] == ">":
                                    count += 1
                                    id_ = line[1:].rstrip(" \t\n\r")
                                    connection.execute(
                                        "INSERT INTO sequences(id, "
                                        "fileNumber, offset) VALUES (?, ?, ?)",
                                        (id_, fileNumber, fp.tell()),
                                    )
                        finally:
                            fp.close()
                else:
                    with open(filename) as fp:
                        offset = 0
                        for line in fp:
                            offset += len(line)
                            if line[0] == ">":
                                count += 1
                                id_ = line[1:].rstrip(" \t\n\r")
                                connection.execute(
                                    "INSERT INTO sequences(id, fileNumber, "
                                    "offset) VALUES (?, ?, ?)",
                                    (id_, fileNumber, offset),
                                )
        except sqlite3.IntegrityError as e:
            if str(e).find("UNIQUE constraint failed") > -1:
                original = self._find(id_)
                if original is None:
                    # The id must have appeared twice in the current file,
                    # because we could not look it up in the database
                    # (i.e., it was INSERTed but not committed).
                    raise ValueError(
                        "FASTA sequence id '%s' found twice in file '%s'."
                        % (id_, filename)
                    )
                else:
                    origFilename, _ = original
                    raise ValueError(
                        "FASTA sequence id '%s', found in file '%s', was "
                        "previously added from file '%s'."
                        % (id_, filename, origFilename)
                    )
            else:
                raise
        else:
            return count

    def _find(self, id_: str) -> Optional[tuple[Optional[str], int]]:
        """
        Find the filename and offset of a sequence, given its id.

        @param id_: A C{str} sequence id.
        @return: A 2-tuple, containing the C{str} file name and C{int} offset
            within that file of the sequence.
        """
        cur = self._connection.cursor()
        cur.execute("SELECT fileNumber, offset FROM sequences WHERE id = ?", (id_,))
        row = cur.fetchone()
        if row is None:
            return None
        else:
            return self._getFilename(row[0]), row[1]

    def __getitem__(self, id_: str) -> Read:
        """
        Return a read, given its id.

        @param id_: A C{str} sequence id.
        @raise KeyError: If C{id_} is not a known sequence.
        @return: A read of our read class.
        """
        location = self._find(id_)
        if location is None:
            raise KeyError("Unknown sequence: %r" % id_)
        else:
            filename, offset = location
            # If a FASTA directory was provided, look for the FASTA files
            # there, otherwise use the filename that was given to addFile.
            assert filename
            if self._fastaDirectory:
                filename = os.path.join(
                    self._fastaDirectory, os.path.basename(filename)
                )

            endswith = filename.lower().endswith
            if endswith(".bgz") or endswith(".gz"):
                opener = bgzf.open
            else:
                opener = open

            sequence = ""
            with opener(filename) as fp:
                fp.seek(offset)
                while True:
                    line = fp.readline()
                    if not line:
                        # EOF
                        break
                    elif line[0] == ">":
                        # We found the next sequence identifier.
                        break
                    else:
                        sequence += line.rstrip("\n\r")

            return self._readClass(id_, sequence)

    def close(self) -> None:
        self._connection.close()



from typing import Iterator, Union
from collections import defaultdict, Counter

from dark.utils import median
from dark.filter import ReadSetFilter, TitleFilter
from dark.hsp import HSP
from dark.intervals import ReadIntervals
from dark.reads import Read


def titleCounts(readsAlignments):
    """
    Count the number of times each title in a readsAlignments instance is
    matched. This is useful for rapidly discovering what titles were matched
    and with what frequency.

    @param readsAlignments: A L{dark.alignments.ReadsAlignments} instance.
    @return: A C{dict} whose keys are titles and whose values are the integer
        counts of the number of reads that matched that title.
    """

    titles = defaultdict(int)
    for readAlignments in readsAlignments:
        for alignment in readAlignments:
            titles[alignment.subjectTitle] += 1
    return titles


class TitleAlignment:
    """
    Hold information about a read's HSPs for a title alignment.

    @param read: The C{Read} that aligned.
    @param hsps: A C{list} of L{dark.hsp.HSP} (or subclass) instances.
    """

    def __init__(self, read: Read, hsps: list[HSP]):
        self.read = read
        self.hsps = hsps

    def toDict(self) -> dict[str, Union[list[dict], dict]]:
        """
        Get information about a title alignment as a dictionary.

        @return: A C{dict} representation of the title aligment.
        """
        return {
            "hsps": [hsp.toDict() for hsp in self.hsps],
            "read": self.read.toDict(),
        }


class TitleAlignments(list):
    """
    Holds information about a list of alignments against a sequence.

    @param subjectTitle: The C{str} title of the sequence the read matched
        against.
    @param subjectLength: The C{int} length of the sequence the read matched
        against.
    """

    def __init__(self, subjectTitle: str, subjectLength: int) -> None:
        self.subjectTitle = subjectTitle
        self.subjectLength = subjectLength

    def addAlignment(self, alignment: TitleAlignment) -> None:
        """
        Add an alignment to the list of alignments that matched this title.

        @param alignment: A L{TitleAlignment} instance.
        """
        self.append(alignment)

    def reads(self) -> Iterator[Read]:
        """
        Find the set of reads matching this title.

        @return: A generator that yields C{dark.reads.Read} instances (or one
            of its subclasses).
        """
        return (alignment.read for alignment in self)

    def readCount(self) -> int:
        """
        Find out how many reads aligned to this title.

        @return: The C{int} number of reads that aligned to this title.
        """
        return len(self)

    def hspCount(self) -> int:
        """
        How many HSPs were there in total for all the alignments to a title.

        @return: The C{int} number of HSPs for the alignments to this title.
        """
        return sum(len(alignment.hsps) for alignment in self)

    def readIds(self) -> set[str]:
        """
        Find the set of read ids that matched the title.

        @return: A C{set} of read ids that aligned to this title.
        """
        return set(alignment.read.id for alignment in self)

    def hsps(self) -> Iterator[HSP]:
        """
        Get all HSPs for the alignments to a title.

        @return: A generator yielding L{dark.hsp.HSP} instances.
        """
        return (hsp for alignment in self for hsp in alignment.hsps)

    def bestHsp(self) -> HSP:
        """
        Find the HSP with the best score.

        @raise ValueError: If there are no HSPs.
        @return: The C{dark.hsp.HSP} instance (or a subclass) with the best
        score.
        """
        return max(hsp for hsp in self.hsps())

    def worstHsp(self) -> HSP:
        """
        Find the HSP with the worst score.

        @raise ValueError: If there are no HSPs.
        @return: The C{dark.hsp.HSP} instance (or a subclass) with the worst
        score.
        """
        return min(hsp for hsp in self.hsps())

    def hasScoreBetterThan(self, score: float) -> bool:
        """
        Is there an HSP with a score better than a given value?

        @param score: A C{float} score.
        @return: A C{bool}, C{True} if there is at least one HSP in the
        alignments for this title with a score better than C{score}.
        """
        # Note: Do not assume that HSPs in an alignment are sorted in
        # decreasing order (as they are in BLAST output). If we could
        # assume that, we could just check the first HSP in each alignment.
        for hsp in self.hsps():
            if hsp.betterThan(score):
                return True
        return False

    def medianScore(self) -> float:
        """
        Find the median score for the HSPs in the alignments that match
        this title.

        @raise ValueError: If there are no HSPs.
        @return: The C{float} median score of HSPs in alignments matching the
            title.
        """
        return median([hsp.score.score for hsp in self.hsps()])

    def coverage(self) -> float:
        """
        Get the fraction of this title sequence that is matched by its reads.

        @return: The C{float} fraction of the title sequence matched by its
            reads.
        """
        intervals = ReadIntervals(self.subjectLength)
        for hsp in self.hsps():
            intervals.add(hsp.subjectStart, hsp.subjectEnd)
        return intervals.coverage()

    def coverageCounts(self):
        """
        For each location in the title sequence, return a count of how many
        times that location is covered by a read.
        """
        intervals = ReadIntervals(self.subjectLength)
        for hsp in self.hsps():
            intervals.add(hsp.subjectStart, hsp.subjectEnd)
        return intervals.coverageCounts()

    def coverageInfo(self):
        """
        Return information about the bases found at each location in our title
        sequence.

        @return: A C{dict} whose keys are C{int} subject offsets and whose
            values are unsorted lists of (score, base) 2-tuples, giving all the
            bases from reads that matched the subject at subject location,
            along with the bit score of the matching read.
        """
        result = defaultdict(list)

        for titleAlignment in self:
            for hsp in titleAlignment.hsps:
                score = hsp.score.score
                for subjectOffset, base, _ in titleAlignment.read.walkHSP(
                    hsp, includeWhiskers=False
                ):
                    result[subjectOffset].append((score, base))

        return result

    def residueCounts(self, convertCaseTo="upper"):
        """
        Count residue frequencies at all sequence locations matched by reads.

        @param convertCaseTo: A C{str}, 'upper', 'lower', or 'none'.
            If 'none', case will not be converted (both the upper and lower
            case string of a residue will be present in the result if they are
            present in the read - usually due to low complexity masking).
        @return: A C{dict} whose keys are C{int} offsets into the title
            sequence and whose values are C{Counters} with the residue as keys
            and the count of that residue at that location as values.
        """
        if convertCaseTo == "none":

            def convert(x):
                return x

        elif convertCaseTo == "lower":
            convert = str.lower
        elif convertCaseTo == "upper":
            convert = str.upper
        else:
            raise ValueError("convertCaseTo must be one of 'none', 'lower', or 'upper'")

        counts = defaultdict(Counter)

        for titleAlignment in self:
            read = titleAlignment.read
            for hsp in titleAlignment.hsps:
                for subjectOffset, residue, inMatch in read.walkHSP(hsp):
                    counts[subjectOffset][convert(residue)] += 1

        return counts

    def summary(self):
        """
        Summarize the alignments for this subject.

        @return: A C{dict} with C{str} keys:
            bestScore: The C{float} best score of the matching reads.
            coverage: The C{float} fraction of the subject genome that is
                matched by at least one read.
            hspCount: The C{int} number of hsps that match the subject.
            medianScore: The C{float} median score of the matching reads.
            readCount: The C{int} number of reads that match the subject.
            subjectLength: The C{int} length of the subject.
            subjectTitle: The C{str} title of the subject.
        """
        return {
            "bestScore": self.bestHsp().score.score,
            "coverage": self.coverage(),
            "hspCount": self.hspCount(),
            "medianScore": self.medianScore(),
            "readCount": self.readCount(),
            "subjectLength": self.subjectLength,
            "subjectTitle": self.subjectTitle,
        }

    def toDict(self):
        """
        Get information about the title's alignments as a dictionary.

        @return: A C{dict} representation of the title's aligments.
        """
        return {
            "titleAlignments": [titleAlignment.toDict() for titleAlignment in self],
            "subjectTitle": self.subjectTitle,
            "subjectLength": self.subjectLength,
        }


class TitlesAlignments(dict):
    """
    Holds (as a dictionary) a set of titles, each with its alignments.

    @param readsAlignments: A L{dark.alignments.ReadsAlignments} instance.
    @param scoreClass: A class to hold and compare scores. If C{None},
        the score class from readsAlignments will be used.
    @param readSetFilter: An instance of dark.filter.ReadSetFilter, or C{None}.
        This can be used to pass a previously used title filter for ongoing
        use in filtering.
    """

    def __init__(self, readsAlignments, scoreClass=None, readSetFilter=None):
        dict.__init__(self)
        self.readsAlignments = readsAlignments
        self.scoreClass = scoreClass or readsAlignments.scoreClass
        self.readSetFilter = readSetFilter
        for readAlignments in readsAlignments:
            for alignment in readAlignments:
                title = alignment.subjectTitle
                try:
                    titleAlignments = self[title]
                except KeyError:
                    titleAlignments = self[title] = TitleAlignments(
                        title, alignment.subjectLength
                    )
                titleAlignments.addAlignment(
                    TitleAlignment(readAlignments.read, alignment.hsps)
                )

    def addTitle(self, title, titleAlignments):
        """
        Add a new title to self.

        @param title: A C{str} title.
        @param titleAlignments: An instance of L{TitleAlignments}.
        @raises KeyError: If the title is already present.
        """
        if title in self:
            raise KeyError(
                "Title %r already present in TitlesAlignments instance." % title
            )
        else:
            self[title] = titleAlignments

    def filter(
        self,
        minMatchingReads=None,
        maxMatchingReads=None,
        minMedianScore=None,
        withScoreBetterThan=None,
        minNewReads=None,
        minCoverage=None,
        maxTitles=None,
        sortOn="maxScore",
        titleRegex=None,
        negativeTitleRegex=None,
    ):
        """
        Filter the titles in self.

        @param minMatchingReads: titles that are matched by fewer reads
            are unacceptable.
        @param maxMatchingReads: titles that are matched by more reads
            are unacceptable.
        @param minMedianScore: sequences that are matched with a median
            bit score that is less are unacceptable.
        @param withScoreBetterThan: if the best score for a title is not
            as good as this value, the title is not acceptable.
        @param minNewReads: The C{float} fraction of its reads by which a new
            title's read set must differ from the read sets of all previously
            seen titles in order for this title to be considered acceptably
            different (and therefore interesting).
        @param minCoverage: The C{float} minimum fraction of the title sequence
            that must be matched by at least one read.
        @param maxTitles: A non-negative C{int} maximum number of titles to
            keep. If more titles than this are present, titles will be sorted
            (according to C{sortOn}) and only the best will be retained.
        @param sortOn: A C{str} attribute to sort on, used only if C{maxTitles}
            is not C{None}. See the C{sortTitles} method below for the legal
            values.
        @param titleRegex: A regex that read ids must match.
        @param negativeTitleRegex: A regex that read ids must not match.
        @raise: C{ValueError} if C{maxTitles} is less than zero or the value of
            C{sortOn} is unknown.
        @return: C{self}, with non-matching titles removed.
        """
        # Use a ReadSetFilter only if we're checking that read sets are
        # sufficiently new.
        if minNewReads is None:
            readSetFilter = None
        else:
            if self.readSetFilter is None:
                self.readSetFilter = ReadSetFilter(minNewReads)
            readSetFilter = self.readSetFilter

        if maxTitles is not None and len(self) > maxTitles:
            if maxTitles < 0:
                raise ValueError("maxTitles (%r) cannot be negative." % maxTitles)
            else:
                # There are too many titles. Make a sorted list of them so
                # we loop through them (below) in the desired order and can
                # break when/if we've reached the maximum. We can't just
                # take the first maxTitles titles from the sorted list now,
                # as some of those titles might later be discarded by the
                # filter and then we'd return a result with fewer titles
                # than we should.
                titles = self.sortTitles(sortOn)
        else:
            titles = list(self)

        if titleRegex or negativeTitleRegex:
            titleFilter = TitleFilter(
                positiveRegex=titleRegex, negativeRegex=negativeTitleRegex
            )
        else:
            titleFilter = None

        titlesToKeep = set()

        for title in titles:
            # Test max titles up front, as it may be zero.
            if maxTitles is not None and len(titlesToKeep) == maxTitles:
                break

            # Test positive and negative regexps.
            if titleFilter and titleFilter.accept(title) == TitleFilter.REJECT:
                continue

            titleAlignments = self[title]
            if (
                minMatchingReads is not None
                and titleAlignments.readCount() < minMatchingReads
            ):
                continue

            if (
                maxMatchingReads is not None
                and titleAlignments.readCount() > maxMatchingReads
            ):
                continue

            # To compare the median score with another score, we must
            # convert both values to instances of the score class used in
            # this data set so they can be compared without us needing to
            # know if numerically greater scores are considered better or
            # not.
            if minMedianScore is not None and self.scoreClass(
                titleAlignments.medianScore()
            ) < self.scoreClass(minMedianScore):
                continue

            if (
                withScoreBetterThan is not None
                and not titleAlignments.hasScoreBetterThan(withScoreBetterThan)
            ):
                continue

            if minCoverage is not None and titleAlignments.coverage() < minCoverage:
                continue

            if readSetFilter and not readSetFilter.accept(title, titleAlignments):
                continue

            titlesToKeep.add(title)

        # Filter self.
        for title in list(self):
            if title not in titlesToKeep:
                del self[title]

        return self

    def hsps(self):
        """
        Get all HSPs for all the alignments for all titles.

        @return: A generator yielding L{dark.hsp.HSP} instances.
        """
        return (
            hsp
            for titleAlignments in self.values()
            for alignment in titleAlignments
            for hsp in alignment.hsps
        )

    def sortTitles(self, by):
        """
        Sort titles by a given attribute and then by title.

        @param by: A C{str}, one of 'length', 'maxScore', 'medianScore',
            'readCount', or 'title'.
        @raise ValueError: If an unknown C{by} value is given.
        @return: A sorted C{list} of titles.
        """
        # First sort titles by the secondary key, which is always the title.
        titles = sorted(iter(self))

        # Then sort on the primary key (if any).
        if by == "length":
            return sorted(
                titles, reverse=True, key=lambda title: self[title].subjectLength
            )
        if by == "maxScore":
            return sorted(titles, reverse=True, key=lambda title: self[title].bestHsp())
        if by == "medianScore":
            return sorted(
                titles,
                reverse=True,
                key=lambda title: self.scoreClass(self[title].medianScore()),
            )
        if by == "readCount":
            return sorted(
                titles, reverse=True, key=lambda title: self[title].readCount()
            )
        if by == "title":
            return titles

        raise ValueError(
            'Sort attribute must be one of "length", "maxScore", '
            '"medianScore", "readCount", "title".'
        )

    def summary(self, sortOn=None):
        """
        Summarize all the alignments for this title.

        @param sortOn: A C{str} attribute to sort titles on. One of 'length',
            'maxScore', 'medianScore', 'readCount', or 'title'.
        @raise ValueError: If an unknown C{sortOn} value is given.
        @return: A generator that yields C{dict} instances as produced by
            C{TitleAlignments} (see class earlier in this file), sorted by
            C{sortOn}.
        """
        titles = self if sortOn is None else self.sortTitles(sortOn)

        for title in titles:
            yield self[title].summary()

    def tabSeparatedSummary(self, sortOn=None):
        """
        Summarize all the alignments for this title as multi-line string with
        TAB-separated values on each line.

        @param sortOn: A C{str} attribute to sort titles on. One of 'length',
            'maxScore', 'medianScore', 'readCount', or 'title'.
        @raise ValueError: If an unknown C{sortOn} value is given.
        @return: A newline-separated C{str}, each line with a summary of a
            title. Each summary line is TAB-separated.
        """
        # The order of the fields returned here is somewhat arbitrary. The
        # subject titles are last because they are so variable in length.
        # Putting them last makes it more likely that the initial columns in
        # printed output will be easier to read down.
        #
        # Note that post-processing scripts will be relying on the field
        # ordering here.  So you can't just add fields. It's probably safe
        # to add them at the end, but be careful / think.
        #
        # A TAB-separated file can easily be read by awk using e.g.,
        # awk 'BEGIN {FS = "\t"} ...'

        result = []
        for titleSummary in self.summary(sortOn):
            result.append(
                "\t".join(
                    [
                        "%(coverage)f",
                        "%(medianScore)f",
                        "%(bestScore)f",
                        "%(readCount)d",
                        "%(hspCount)d",
                        "%(subjectLength)d",
                        "%(subjectTitle)s",
                    ]
                )
                % titleSummary
            )
        return "\n".join(result)

    def toDict(self):
        """
        Get information about the titles alignments as a dictionary.

        @return: A C{dict} representation of the titles aligments.
        """
        return {
            "scoreClass": self.scoreClass.__name__,
            "titles": dict(
                (title, titleAlignments.toDict())
                for title, titleAlignments in self.items()
            ),
        }



#!/usr/bin/env python

import sys

from dark.filter import (
    addFASTAFilteringCommandLineOptions,
    parseFASTAFilteringCommandLineOptions,
    addFASTAEditingCommandLineOptions,
    parseFASTAEditingCommandLineOptions,
)
from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Given FASTA on stdin and a set of filtering criteria "
            "write filtered FASTA to stdout."
        ),
    )

    parser.add_argument(
        "--quiet",
        action="store_true",
        default=False,
        help=("If True, do not print the final sequence summary."),
    )

    parser.add_argument(
        "--saveAs",
        choices=("fasta", "fastq", "fasta-ss"),
        help=(
            "The output format. The default is to match the input format, "
            "so there is usually no need to specify this option. It can be "
            "used to force conversion from FASTQ to FASTA"
        ),
    )

    parser.add_argument(
        "--checkResultCount",
        type=int,
        help=(
            "The number of reads expected in the output. If this number is "
            "not seen, the script exits with status 1 and an error "
            "message is printed unless --quiet was used."
        ),
    )

    addFASTACommandLineOptions(parser)
    addFASTAFilteringCommandLineOptions(parser)
    addFASTAEditingCommandLineOptions(parser)

    args = parser.parse_args()

    # Note that we must call parseFASTACommandLineOptions here before we
    # examine args.fasta in the following code. That's because
    # parseFASTACommandLineOptions sets args.fasta to be True if no format
    # is given.
    reads = parseFASTAEditingCommandLineOptions(
        args,
        parseFASTAFilteringCommandLineOptions(args, parseFASTACommandLineOptions(args)),
    )

    saveAs = (
        args.saveAs
        or (args.fasta and "fasta")
        or (args.fastq and "fastq")
        or (args.fasta_ss and "fasta-ss")
    )

    # Check for incompatible read/write formats. We can't write FASTQ
    # unless we have FASTQ on input (else we won't have quality information),
    # and we can't write PDB FASTA with secondary structure information
    # unless we have that on input.
    if saveAs == "fastq" and not args.fastq:
        raise ValueError(
            "You have specified --saveAs fastq without using --fastq "
            "to indicate that the input is FASTQ. Please be explicit."
        )
    elif saveAs == "fasta-ss" and not args.fasta_ss:
        raise ValueError(
            "You have specified --saveAs fasta-ss without using --fasta-ss "
            "to indicate that the input is PDB FASTA. Please be explicit."
        )

    write = sys.stdout.write
    kept = 0
    for kept, read in enumerate(reads, start=1):
        write(read.toString(format_=saveAs))

    total = reads.unfilteredLength()

    if not args.quiet:
        print(
            "Read %d sequence%s, kept %d (%.2f%%)."
            % (
                total,
                "" if total == 1 else "s",
                kept,
                0.0 if total == 0 else kept / total * 100.0,
            ),
            file=sys.stderr,
        )

    if args.checkResultCount is not None:
        if kept != args.checkResultCount:
            print(
                "Did not write the expected %d sequence%s (wrote %d)."
                % (
                    args.checkResultCount,
                    "" if args.checkResultCount == 1 else "s",
                    kept,
                ),
                file=sys.stderr,
            )
            sys.exit(1)



#!/usr/bin/env python

import sys
import argparse

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=(
        "Find sequences that have a specific property (as detected "
        'by the "relabel" function in the file specified by '
        "--relabelFunctionFile) and print a TAB-separated list of "
        "renamings to standard output."
    ),
)

parser.add_argument(
    "--relabelFunctionFile",
    required=True,
    help=(
        'A file containing a Python function named "relabel" that takes a '
        "list of reads and returns a dictionary of renamings."
    ),
)

addFASTACommandLineOptions(parser)
args = parser.parse_args()
reads = parseFASTACommandLineOptions(args)

relabel = origRelabel = object()

try:
    exec(open(args.relabelFunctionFile).read())
except Exception as e:
    print(
        "Could not execute Python in %s: %s" % (args.relabelFunctionFile, e),
        file=sys.stderr,
    )
    sys.exit(1)

if relabel is origRelabel:
    print(
        'The Python file %s did not define a function called "relabel".'
        % args.relabelFunctionFile,
        file=sys.stderr,
    )
    sys.exit(2)

for old, new in relabel(reads).items():
    print("%s\t%s" % (old, new))



#!/usr/bin/env python

import sys
import os
import argparse

from dark.reads import (
    addFASTACommandLineOptions,
    parseFASTACommandLineOptions,
    getUnambiguousBases,
)

parser = argparse.ArgumentParser(
    description=(
        "Given FASTA on standard input, and bases to look for, write the "
        "1-based indices of where the bases occur in all sequences to "
        "standard output. If standard output is a terminal, these will be "
        "space separated, else newline separated. Use --any to print "
        "indices that match any sequence."
    ),
    epilog=(
        "This can be used to find all columns of a FASTA multiple "
        "sequence alignment that contain gaps, ambiguous nucleotides or "
        "AAs, or to find all columns that do not contain such things. "
        "Note that if sequences are of uneven lengths and --any is not "
        "specified, only indices up to the length of the shortest input "
        "sequence can be printed (i.e., there is a strict interpretation "
        "of all sequences needing to have a matching base at an index: "
        "if the index does not exist in even one sequence, then no base "
        "can occur in all sequences at that index)."
    ),
)

parser.add_argument(
    "--bases",
    help=(
        "The sequence bases whose indices should be printed. If not "
        "specified, this will be the defined set of bases for the input "
        'sequence type (i.e., "ACGT" for DNA). This will have the effect of '
        "printing the indices for which any sequence has an ambiguous or "
        "missing base."
    ),
)

parser.add_argument(
    "--matchCase",
    default=False,
    action="store_true",
    help="If specified, sequence case will be considered in matching.",
)

parser.add_argument(
    "--any",
    default=False,
    action="store_true",
    help=(
        "If specified, print indices of bases that match any sequence, "
        "otherwise (the default) indices are only printed if they match "
        "all sequences."
    ),
)

addFASTACommandLineOptions(parser)
args = parser.parse_args()
reads = parseFASTACommandLineOptions(args)

if args.bases is None:
    # No target bases were given. Use the set of unambiguous bases for
    # the read type. Unless --any has been given, this will print
    # indices in which no ambiguous bases or gaps appear in any
    # sequence.
    targets = getUnambiguousBases[args.readClass]
else:
    targets = set(args.bases)

indices = reads.indicesMatching(targets, args.matchCase, args.any)
nIndices = len(indices)

if nIndices:
    separator = " " if os.isatty(1) else "\n"
    # Add one to indices to maximize happy humans.
    print(separator.join(map(lambda x: str(x + 1), sorted(indices))))

print(
    "Found %d %s where %s a base from the set {%s}."
    % (
        nIndices,
        "index" if nIndices == 1 else "indices",
        "any sequence has" if args.any else "all sequences have",
        ", ".join(sorted(targets)),
    ),
    file=sys.stderr,
)



#!/usr/bin/env python

import sys
import argparse
from math import log10
from pathlib import Path
from typing import Optional

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions, DNARead


def getArgsAndReads():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Split a FASTA alignment into equal-sized windows.",
    )

    addFASTACommandLineOptions(parser)

    parser.add_argument(
        "--base",
        default="window-",
        help=(
            "The base output filename. If you want output to be placed in a "
            "sub-directory, specify the sub-directory in the value. Required "
            "non-existent intermediate directories (if any) will be created."
        ),
    )

    parser.add_argument("--window", type=int, default=100, help="The window size.")

    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Write processing information to standard error.",
    )

    parser.add_argument(
        "--zeroBased",
        action="store_false",
        dest="oneBased",
        help=(
            "The --startOffset and offsets used in filenames will be zero-based. If "
            "not given, offsets are one-based."
        ),
    )

    parser.add_argument(
        "--minWindow",
        type=int,
        help=(
            "If the length of the final windowed sequence is less than this value, "
            "do not write a file for it."
        ),
    )

    parser.add_argument(
        "--maxWindows",
        type=int,
        help="The maximum number of windowed subsequences to extract.",
    )

    parser.add_argument(
        "--rotate",
        type=int,
        default=0,
        metavar="N",
        help=(
            "Rotate sequences left (if N is negative) or right (if positive) before "
            "extracting sub-sequence windows. Output filenames will be adjusted to "
            "match the rotation amount and the sequence length. For example, if the "
            "window size is 200, a --rotate value of -100 is given, and the sequences "
            "have length 2000, the first output file will have a name like "
            "window-1901-0100.fasta (or window-1900-0100.fasta if --zeroBased is "
            "used) and the final name will be window-1701-1900.fasta. This option is "
            "useful for making windows that match the location of genome features, "
            "especially in the case of a circular genome (e.g., for hepatitis B virus)."
        ),
    )

    parser.add_argument(
        "--startOffset",
        type=int,
        help=(
            "The zero-based genome offset to start windowing from (default value 0). "
            "Use --oneBased if you would prefer to use a one-based offset (in which "
            "case the default value is 1 and indices in the output files will also be "
            "one-based)."
        ),
    )

    parser.add_argument(
        "--invalidChars",
        default="N?-",
        help=(
            "Sequence characters (case insensitive) to consider invalid for the "
            "purposes of tree-making. If a windowed sequence for a sample consists "
            "entirely of these characters, no sequence will be written to the "
            "output for that window of that sample."
        ),
    )

    args = parser.parse_args()

    if args.maxWindows is not None and args.maxWindows < 1:
        exit("The value given to --maxWindows cannot be less than one.")

    if args.startOffset is None:
        args.startOffset = int(args.oneBased)

    if args.oneBased:
        if args.startOffset < 1:
            exit(
                "When using --oneBased, the --startOffset value must be greater than "
                "zero."
            )
    else:
        if args.startOffset < 0:
            exit("The --startOffset value cannot be less than zero.")

    reads = parseFASTACommandLineOptions(args)

    if rotate := args.rotate:
        result = [read.rotate(rotate) for read in reads]
    else:
        result = list(reads)

    return args, result


def checkReads(reads: list[DNARead]) -> int:
    """
    Check all reads have the same length and return that length.

    @param reads: A C{list} of C{DNARead} instances.
    @return: The C{int} length of the reads.
    """
    assert reads, "Empty FASTA input!"
    length = len(reads[0])
    assert length, "The first input sequence has length zero!"

    for line, read in enumerate(reads[1:], start=2):
        if len(read) != length:
            exit(
                f"Sequence with id {read.id!r} on input line {line} has length "
                f"{len(read)} which is not equal to the previous input sequence "
                f"length ({length})."
            )

    return length


def writeFiles(
    reads: list[DNARead],
    window: int,
    maxWindows: int,
    minWindow: Optional[int],
    startOffset: int,
    rotate: int,
    oneBased: bool,
    length: int,
    base: str,
    invalidChars: str,
    verbose: bool = False,
) -> None:
    """
    Write windowed (i.e., sub-alignment) FASTA files.

    @param reads: A C{list} of C{DNARead} instances.
    @param window: The C{int} window size.
    @param maxWindows: The C{int} maximum number of windows to produce.
    @param minWindow: The C{int} minimum window size or C{None} if there is no minimum.
    @param startOffset: The C{int} starting offset in the genome, interpreted
        according to C{zeroBased}.
    @param rotate: The C{int} amount the reads were rotated (zero means they were not
        rotated at all). The reads have already been rotated, so this argument is only
        needed to adjust the output filenames.
    @param oneBased: If C{True}, the passed C{startOffset} and the offsets
        used in output file names will be one-based. Else, zero-based.
    @param length: The C{int} length of all sequences in C{reads}.
    @param base: The C{str} basename of the output files.
    @param invalidChars: The C{str} (uppercase) sequence characters to ignore
        when deciding if a windowed sequence is valid.
    @param verbose: If C{True} write processing information to C{sys.stderr}.
    """
    width = int(log10(length)) + 1

    # Adjust startOffset to be zero-based.
    startOffset -= oneBased

    for windowCount, start in enumerate(range(startOffset, length, window)):

        if windowCount == maxWindows:
            break

        end = start + window
        if end > length:
            end = length

        # displayStart is the start value we print in output messages and use in
        # filenames. It is the zero-based window start offset, including any rotation
        # that has already been performed on the reads, adjusted upwards by one if we
        # have been told to use one-based values. displayEnd is similar (but does not
        # need to be adjusted by one due to Python's zero-indexing and exclusion of the
        # final index in ranges).
        displayStart = start + rotate
        displayEnd = displayStart + (end - start)

        if displayStart < 0:
            displayStart += length
        elif displayStart >= length:
            displayStart -= length

        displayStart += oneBased

        if displayEnd < 0:
            displayEnd += length
        elif displayEnd > length:
            displayEnd -= length

        tooShort = minWindow is not None and end - start < minWindow

        filename = Path(f"{base}{displayStart:0{width}d}-{displayEnd:0{width}d}.fasta")

        if not tooShort and not filename.parent.exists():
            filename.parent.mkdir(parents=True)

        if verbose:
            print(f"Processing window {displayStart}-{displayEnd}:", file=sys.stderr)

        if tooShort:
            # Check that we are at the end of the genome.
            assert end == length
            if verbose:
                print(
                    f"  Final window (length {end - start}) shorter than the minimum window size "
                    f"({minWindow}). Skipping.",
                    file=sys.stderr,
                )
            break

        count = 0
        with open(filename, "w") as fp:
            for read in reads:
                new = read[start:end]
                sequence = new.sequence.upper()
                for char in invalidChars:
                    sequence = sequence.replace(char, "")
                if sequence:
                    count += 1
                    print(new.toString(), end="", file=fp)
                else:
                    if verbose:
                        print(
                            f"  Window {displayStart}-{displayEnd} for sequence "
                            f"{read.id!r} is all {invalidChars!r} characters. Skipping.",
                            file=sys.stderr,
                        )

        if count:
            if verbose:
                print(
                    f"  Wrote {count} window {displayStart}-{displayEnd} "
                    f"sequence{'s' if count else ''} to {str(filename)!r}.",
                    file=sys.stderr,
                )
        else:
            if verbose:
                print(
                    f"  No sequences with valid characters were found for window "
                    f"{displayStart}-{displayEnd}.",
                    file=sys.stderr,
                )
            filename.unlink()


def main():
    args, reads = getArgsAndReads()
    length = checkReads(reads)
    writeFiles(
        reads,
        args.window,
        args.maxWindows,
        args.minWindow,
        args.startOffset,
        args.rotate,
        args.oneBased,
        length,
        args.base,
        args.invalidChars.upper(),
        args.verbose,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import sys
import argparse
from itertools import chain

from dark.civ.proteins import SqliteIndex
from dark.errors import NoSuchGenomeError
from dark.filter import (
    addFASTAFilteringCommandLineOptions,
    parseFASTAFilteringCommandLineOptions,
)
from dark.genbank import GenomeRanges
from dark.genomes import GenomeProteinInfo
from dark.reads import Reads
from dark.sam import SAMFilter, samReferences
from dark.utils import pct


def summarize(gpi, sortOn, minReadOffsetCount):
    """
    Print a summary of the genome proteins.

    @param gpi: A C{GenomeProteinInfo} instance.
    @param sortOn: How to sort proteins for output. One of 'coverage',
        'depth', 'name', 'offset', or 'readCount'.
    @param minReadOffsetCount: The minimum number of reads offsets that must
        overlap a protein for the read to be considered as sufficiently
        intersecting the protein.
    """
    genome = gpi.genome

    print("Summary of %s (%s):" % (genome["name"], genome["accession"]))
    print("  Length: %d" % genome["length"])
    print("  Protein count: %d" % genome["proteinCount"])
    print("  Total protein offsets: %s" % (pct(len(gpi.offsets), genome["length"])))

    if gpi.samFiles:
        print("  SAM files analyzed: %d" % len(gpi.samFiles))
        for i, filename in enumerate(gpi.samFiles, start=1):
            print("    %d: %s" % (i, filename))
    else:
        return

    print("  Whole genome coverage (not just proteins):")
    print("    Reads matching genome: %d" % len(gpi.readIdsMatchingGenome))
    print(
        "    Covered genome offsets: %s"
        % (pct(len(gpi.coveredOffsetCount), genome["length"]))
    )
    print(
        "    Average depth across genome: %.3f"
        % (sum(gpi.coveredOffsetCount.values()) / genome["length"])
    )

    coveredProteinOffsetCount = coveredProteinBasesCount = 0
    for offset in gpi.offsets:
        coveredProteinOffsetCount += bool(gpi.coveredOffsetCount[offset])
        coveredProteinBasesCount += gpi.coveredOffsetCount[offset]

    print("  Total protein coverage (irrespective of minReadOffsetCount):")
    print("    Reads matching proteins: %d" % len(gpi.readIdsForAllProteins()))
    print(
        "    Proteins with any coverage: %s"
        % pct(len(gpi.coveredProteins), genome["proteinCount"])
    )
    print(
        "    Covered protein offsets: %s"
        % (pct(coveredProteinOffsetCount, len(gpi.offsets)))
    )
    print(
        "    Average depth across proteins: %.3f"
        % (coveredProteinBasesCount / len(gpi.offsets))
    )

    if sortOn == "name":

        def key(proteinAccession):
            return gpi.proteins[proteinAccession]["name"]

        reverse = False
    elif sortOn == "offset":

        def key(proteinAccession):
            return GenomeRanges(gpi.proteins[proteinAccession]["offsets"]).ranges[0][0]

        reverse = False
    elif sortOn == "readCount":

        def key(proteinAccession):
            coverage = gpi.proteinCoverageInfo(proteinAccession)
            return len(coverage["readIds"])

        reverse = True
    elif sortOn == "coverage":

        def key(proteinAccession):
            coverage = gpi.proteinCoverageInfo(proteinAccession)
            return coverage["coveredOffsets"] / coverage["ntLength"]

        reverse = True
    elif sortOn == "depth":

        def key(proteinAccession):
            coverage = gpi.proteinCoverageInfo(proteinAccession)
            return coverage["totalBases"] / coverage["ntLength"]

        reverse = True

    if minReadOffsetCount is None:
        print("  Proteins covered (no minReadOffsetCount):")
    else:
        print("  Proteins covered (minReadOffsetCount=%d):" % minReadOffsetCount)

    proteinCount = 0
    for proteinAccession in sorted(gpi.coveredProteins, key=key, reverse=reverse):
        protein = gpi.proteins[proteinAccession]

        coverage = gpi.proteinCoverageInfo(proteinAccession, minReadOffsetCount)

        readCount = len(coverage["readIds"])

        if readCount:
            proteinCount += 1

            print(
                "    %d: %s (%d AA, %d nt with stop codon, %s)"
                % (
                    proteinCount,
                    protein["product"],
                    protein["length"],
                    protein["length"] * 3 + 3,
                    protein["accession"],
                )
            )

            print("      Read count: %d" % readCount)

            print(
                "      Covered offsets: %s"
                % (pct(coverage["coveredOffsets"], coverage["ntLength"]))
            )

            print(
                "      Average depth: %.3f"
                % (coverage["totalBases"] / coverage["ntLength"])
            )

            print("      Offsets: %s" % protein["offsets"])

    print(
        "  Proteins matched: %s (sorted by %s):"
        % (pct(proteinCount, genome["proteinCount"]), sortOn)
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Print SAM/BAM file protein match statistics.",
    )

    parser.add_argument(
        "--proteinGenomeDatabase",
        required=True,
        help=(
            "The filename of an Sqlite3 database holding protein and "
            "genome information, as built by make-protein-database.py"
        ),
    )

    parser.add_argument(
        "--progress",
        default=False,
        action="store_true",
        help="Print progress info to standard error.",
    )

    parser.add_argument(
        "--sortOn",
        default="readCount",
        choices=("coverage", "depth", "name", "offset", "readCount"),
        help="How to sort proteins for output.",
    )

    parser.add_argument(
        "--minReadOffsetCount",
        type=int,
        help=(
            "The minimum number of reads offsets that must overlap a "
            "protein for the read to be considered as sufficiently "
            "intersecting the protein. Use this to prevent reads that "
            "just overlap the protein in a very small number offsets "
            "from being counted."
        ),
    )

    parser.add_argument(
        "--skipTranslationChecks",
        dest="checkTranslations",
        action="store_false",
        default=True,
        help=(
            "Skip the sanity check that database protein sequences can all "
            "be translated from the database genome sequence."
        ),
    )

    addFASTAFilteringCommandLineOptions(parser)
    SAMFilter.addFilteringOptions(
        parser,
        samfileIsPositional=False,
        samfileAction="append",
        samfileNargs="*",
        samfileRequired=False,
    )

    args = parser.parse_args()

    samfiles = list(chain.from_iterable(args.samfile)) if args.samfile else []

    if samfiles:
        if args.referenceId:
            referenceIds = args.referenceId
        else:
            # If all SAM files have just one reference and they're all the
            # same, use that. Else complain.
            referenceIds = set()
            for filename in samfiles:
                referenceIds.update(samReferences(filename))

            if len(referenceIds) != 1:
                print(
                    "No reference id(s) specified with --referenceId, and "
                    "the given SAM/BAM files do not contain exactly one "
                    "(identical) reference. Please use --referenceId"
                )
                sys.exit(1)
    else:
        if args.referenceId:
            referenceIds = args.referenceId
        else:
            print("No reference id(s) specified with --referenceId.")
            sys.exit(1)

    # We don't have a file of reads, we just want a read filter that we
    # can use to filter the SAM file query sequences.
    reads = parseFASTAFilteringCommandLineOptions(args, Reads())
    samFilter = SAMFilter.parseFilteringOptions(args, reads.filterRead)
    filterAlignment = samFilter.filterAlignment

    proteinGenomeDB = SqliteIndex(args.proteinGenomeDatabase)

    for referenceId in referenceIds:
        try:
            gpInfo = GenomeProteinInfo(
                referenceId, proteinGenomeDB, checkTranslations=args.checkTranslations
            )
        except NoSuchGenomeError:
            print(
                "Reference %r not found in genome database. Ignoring." % referenceId,
                file=sys.stderr,
            )
        else:
            if samfiles:
                if args.progress:
                    print(
                        "Processing %d SAM file%s for matches with %r:"
                        % (
                            len(samfiles),
                            "" if len(samfiles) == 1 else "s",
                            referenceId,
                        ),
                        file=sys.stderr,
                    )
                for i, filename in enumerate(samfiles, start=1):
                    if args.progress:
                        print("  %d: %s" % (i, filename), file=sys.stderr)
                    gpInfo.addSAM(filename, filterAlignment)

            summarize(gpInfo, args.sortOn, args.minReadOffsetCount)



#!/usr/bin/env python

from re import compile
import sys

from dark.entrez import getSequence


def main(gi, ranges):
    """
    Print the features of the genbank entry given by gi. If ranges is
    non-emtpy, only print features that include the ranges.

    gi: either a hit from a BLAST record, in the form
        'gi|63148399|gb|DQ011818.1|' or a gi number (63148399 in this example).
    ranges: a possibly empty list of ranges to print information for. Each
        range is a non-descending (start, end) pair of integers.
    """
    # TODO: Make it so we can pass a 'db' argument to getSequence.
    record = getSequence(gi)

    if record is None:
        print("Looks like you're offline.")
        sys.exit(3)
    else:
        printed = set()
        if ranges:
            for start, end in ranges:
                for index, feature in enumerate(record.features):
                    if (
                        start < int(feature.location.end)
                        and end > int(feature.location.start)
                        and index not in printed
                    ):
                        print(feature)
                        printed.add(index)
        else:
            # Print all features.
            for feature in record.features:
                print(feature)


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print(
            "Usage: %s gi-number [offset1, offset2, ...]" % sys.argv[0], file=sys.stderr
        )
        sys.exit(1)

    rangeRegex = compile(r"^(\d+)(?:-(\d+))?$")
    ranges = []
    for arg in sys.argv[2:]:
        match = rangeRegex.match(arg)
        if match:
            start, end = match.groups()
            start = int(start)
            if end is None:
                end = start
            else:
                end = int(end)
            if start > end:
                start, end = end, start
            ranges.append((start, end))
        else:
            print(
                (
                    "Illegal argument %r. Ranges must single numbers or "
                    "number-number." % arg
                ),
                file=sys.stderr,
            )
            sys.exit(2)

    main(sys.argv[1], ranges)



#!/usr/bin/env python

from random import choice
from operator import itemgetter

from dark.aaVars import CODONS
from dark.reads import Read, addFASTACommandLineOptions, parseFASTACommandLineOptions


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description=(
            "Given amino acid FASTA on stdin, write FASTA with DNA "
            "sequences to stdout, where the DNA sequences would "
            "translate to the given amino acid seqeunces. The DNA "
            "bases used for each amino acid is either fixed or is "
            "chosen at random from the corresponding codons."
        )
    )

    parser.add_argument(
        "--random",
        action="store_true",
        help=(
            "Use codons chosen at random for each amino acid. Otherwise a "
            "fixed codon will be used for each amino acid so that the "
            "output for a given amino acid sequence is always identical."
        ),
    )

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    reads = parseFASTACommandLineOptions(args)

    getCodon = choice if args.random else itemgetter(0)

    for aaRead in reads:
        dnaRead = Read(
            aaRead.id, "".join(getCodon(CODONS[aa]) for aa in aaRead.sequence)
        )
        print(dnaRead.toString("fasta"), end="")
